from __future__ import annotations

__all__ = (
    "update_action_button",
    "update_action_link",
    "update_checkbox",
    "update_switch",
    "update_checkbox_group",
    "update_radio_buttons",
    "update_date",
    "update_date_range",
    "update_numeric",
    "update_select",
    "update_selectize",
    "update_slider",
    "update_text",
    "update_text_area",
    "update_navs",
)

import json
import re
from datetime import date
from typing import TYPE_CHECKING, Literal, Mapping, Optional, cast, overload

from htmltools import TagChild, TagList, tags
from starlette.requests import Request
from starlette.responses import JSONResponse, Response

from .._docstring import add_example, doc_format, no_example
from .._namespaces import ResolvedId, resolve_id
from .._typing_extensions import NotRequired, TypedDict
from .._utils import drop_none
from ..input_handler import input_handlers
from ..session import require_active_session, session_context
from ..types import ActionButtonValue
from ._input_check_radio import ChoicesArg, _generate_options
from ._input_date import _as_date_attr
from ._input_select import SelectChoicesArg, _normalize_choices, _render_choices
from ._input_slider import SliderStepArg, SliderValueArg, _as_numeric, _slider_type
from ._utils import JSEval, _session_on_flush_send_msg, extract_js_keys

if TYPE_CHECKING:
    from ..session import Session


_note = """
    The input updater functions send a message to the client, telling it to change the
    settings of an input object. The messages are collected and sent after all the
    observers (including outputs) have finished running.

    The syntax of these functions is similar to the functions that created the inputs in
    the first place. For example, :func:`~shiny.ui.input_numeric` and
    :func:`~shiny.ui.update_numeric` take a similar set of arguments.

    Any arguments with ``None`` values will be ignored; they will not result in any
    changes to the input object on the client.

    For :func:`~shiny.ui.update_radio_buttons`, :func:`~shiny.ui.update_checkbox_group`, and
    :func:`~shiny.ui.update_select`, the set of choices can be cleared by using ``choices=[]``.
    Similarly, for these inputs, the selected item can be cleared by using
    `selected=[]`.
"""


# -----------------------------------------------------------------------------
# input_action_button.py
# -----------------------------------------------------------------------------
@add_example()
@doc_format(note=_note)
def update_action_button(
    id: str,
    *,
    label: Optional[str] = None,
    icon: TagChild = None,
    disabled: Optional[bool] = None,
    session: Optional[Session] = None,
) -> None:
    """
    Change the label and/or icon of an action button on the client.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    icon
        An icon to appear inline with the button/link.
    disabled
        If `True`, disable the button making it unclickable; if `False`, the button will
        become enabled and clickable.
    session
        A :class:`~shiny.Session` instance. If not provided, it is inferred via
        :func:`~shiny.session.get_current_session`.

    Note
    ----
    {note}

    See Also
    --------
    * :func:`~shiny.input_action_button`
    """

    session = require_active_session(session)
    # TODO: supporting a TagChild for label would require changes to shiny.js
    # https://github.com/rstudio/shiny/issues/1140
    msg = {
        "label": label,
        "icon": session._process_ui(icon)["html"] if icon else None,
        "disabled": disabled,
    }
    session.send_input_message(id, drop_none(msg))


@add_example(ex_dir="../api-examples/update_action_button")
@doc_format(note=_note)
def update_action_link(
    id: str,
    *,
    label: Optional[str] = None,
    icon: TagChild = None,
    session: Optional[Session] = None,
) -> None:
    """
    Change the label and/or icon of an action link on the client.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    icon
        An icon to appear inline with the button/link.
    session
        A :class:`~shiny.Session` instance. If not provided, it is inferred via
        :func:`~shiny.session.get_current_session`.

    Note
    ----
    {note}

    See Also
    --------
    * :func:`~shiny.input_action_link`
    """

    session = require_active_session(session)
    # TODO: supporting a TagChild for label would require changes to shiny.js
    # https://github.com/rstudio/shiny/issues/1140
    msg = {
        "label": label,
        "icon": session._process_ui(icon)["html"] if icon else None,
    }
    session.send_input_message(id, drop_none(msg))


# -----------------------------------------------------------------------------
# input_task_button.py
# -----------------------------------------------------------------------------
@no_example()
def update_task_button(
    id: str,
    *,
    state: Optional[str] = None,
    session: Optional[Session] = None,
) -> None:
    """
    Change the state of a task button on the client.

    When a task button is clicked, it automatically changes to the "busy" state. This
    function can be used to change the state back to "ready" when the task is complete.

    You can also use this function to change the state to "busy" manually, which will
    prevent the button from automatically resetting to "ready" after a click.

    Parameters
    ----------
    id
        An input id.
    state
        The new state of the button. One of "ready", "busy", or a custom state name
        added via :func:`~shiny.ui.input_task_button`.
    session
        A :class:`~shiny.Session` instance. If not provided, it is inferred via
        :func:`~shiny.session.get_current_session`.
    """

    session = require_active_session(session)

    if state is not None:
        resolved_id = session.ns(id)
        if state != "ready":
            manual_task_reset_buttons.add(resolved_id)
        else:
            manual_task_reset_buttons.discard(resolved_id)

    msg = {"state": state}
    session.send_input_message(id, drop_none(msg))


manual_task_reset_buttons: set[ResolvedId] = set()


@input_handlers.add("bslib.taskbutton")
def _(
    value: dict[str, object], name: ResolvedId, session: Session
) -> ActionButtonValue:
    if value["autoReset"]:

        @session.on_flush
        def callback() -> None:
            # This is input_task_button's auto-reset feature: unless the button has
            # opted out using set_task_button_manual_reset(), we should reset after a
            # flush cycle where a bslib.taskbutton value is seen.
            if name not in manual_task_reset_buttons:
                update_task_button(name, state="ready", session=session)

    return ActionButtonValue(cast(int, value["value"]))


# -----------------------------------------------------------------------------
# input_check_radio.py
# -----------------------------------------------------------------------------
@add_example()
@doc_format(note=_note)
def update_checkbox(
    id: str,
    *,
    label: Optional[str] = None,
    value: Optional[bool] = None,
    session: Optional[Session] = None,
) -> None:
    """
    Change the value of a checkbox input on the client.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    value
        A new value.
    session
        A :class:`~shiny.Session` instance. If not provided, it is inferred via
        :func:`~shiny.session.get_current_session`.

    Note
    ----
    {note}

    See Also
    --------
    * :func:`~shiny.ui.input_checkbox`
    """

    session = require_active_session(session)
    msg = {"label": label, "value": value}
    session.send_input_message(id, drop_none(msg))


@no_example()
@doc_format(note=_note)
def update_switch(
    id: str,
    *,
    label: Optional[str] = None,
    value: Optional[bool] = None,
    session: Optional[Session] = None,
) -> None:
    """
    Change the value of a switch input on the client.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    value
        A new value.
    session
        A :class:`~shiny.Session` instance. If not provided, it is inferred via
        :func:`~shiny.session.get_current_session`.

    Note
    ----
    {note}

    See Also
    --------
    * :func:`~shiny.ui.input_switch`
    """

    session = require_active_session(session)
    msg = {"label": label, "value": value}
    session.send_input_message(id, drop_none(msg))


@add_example()
@doc_format(note=_note)
def update_checkbox_group(
    id: str,
    *,
    label: Optional[str] = None,
    choices: Optional[ChoicesArg] = None,
    selected: Optional[str | list[str] | tuple[str, ...]] = None,
    inline: bool = False,
    session: Optional[Session] = None,
) -> None:
    """
    Change the value of a checkbox group input on the client.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    choices
        Either a list of choices or a dictionary mapping choice values to labels. Note
        that if a dictionary is provided, the keys are used as the (input) values so
        that the dictionary values can hold HTML labels.
    selected
        The values that should be initially selected, if any.
    inline
        If ``True``, the result is displayed inline
    session
        A :class:`~shiny.Session` instance. If not provided, it is inferred via
        :func:`~shiny.session.get_current_session`.

    Note
    ----
    {note}

    See Also
    --------
    * :func:`~shiny.ui.input_checkbox_group`
    """

    _update_choice_input(
        id=resolve_id(id),
        type="checkbox",
        label=label,
        choices=choices,
        selected=selected,
        inline=inline,
        session=session,
    )


@add_example()
@doc_format(note=_note)
def update_radio_buttons(
    id: str,
    *,
    label: Optional[str] = None,
    choices: Optional[ChoicesArg] = None,
    selected: Optional[str] = None,
    inline: bool = False,
    session: Optional[Session] = None,
) -> None:
    """
    Change the value of a radio input on the client.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    choices
        Either a list of choices or a dictionary mapping choice values to labels. Note
        that if a dictionary is provided, the keys are used as the (input) values so
        that the dictionary values can hold HTML labels.
    selected
        The values that should be initially selected, if any.
    inline
        If ``True```, the result is displayed inline
    session
        A :class:`~shiny.Session` instance. If not provided, it is inferred via
        :func:`~shiny.session.get_current_session`.

    Note
    ----
    {note}

    See Also
    --------
    * :func:`~shiny.ui.input_radio_buttons`
    """

    _update_choice_input(
        id=resolve_id(id),
        type="radio",
        label=label,
        choices=choices,
        selected=selected,
        inline=inline,
        session=session,
    )


def _update_choice_input(
    id: str,
    *,
    type: Literal["checkbox", "radio"],
    label: Optional[str] = None,
    choices: Optional[ChoicesArg] = None,
    selected: Optional[str | list[str] | tuple[str, ...]] = None,
    inline: bool = False,
    session: Optional[Session] = None,
) -> None:
    session = require_active_session(session)
    options = None
    if choices is not None:
        # https://github.com/posit-dev/py-shiny/issues/708#issuecomment-1696352934
        with session_context(session):
            resolved_id = resolve_id(id)

        opts = _generate_options(
            id=resolved_id,
            type=type,
            choices=choices,
            selected=selected,
            inline=inline,
        )
        options = session._process_ui(opts)["html"]
    msg = {"label": label, "options": options, "value": selected}
    session.send_input_message(id, drop_none(msg))


# -----------------------------------------------------------------------------
# input_date.py
# -----------------------------------------------------------------------------
@add_example()
@doc_format(note=_note)
def update_date(
    id: str,
    *,
    label: Optional[str] = None,
    value: Optional[date | str] = None,
    min: Optional[date | str] = None,
    max: Optional[date | str] = None,
    session: Optional[Session] = None,
) -> None:
    """
    Change the value of a date input on the client.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    value
        The starting date. Either a `date()` object, or a string in yyyy-mm-dd format.
        If ``None`` (the default), will use the current date in the client's time zone.
    min
        The minimum allowed value.
    max
        The maximum allowed value.
    session
        A :class:`~shiny.Session` instance. If not provided, it is inferred via
        :func:`~shiny.session.get_current_session`.

    Note
    ----
    {note}

    See Also
    --------
    * :func:`~shiny.ui.input_date`
    """

    session = require_active_session(session)
    msg = {
        "label": label,
        "value": _as_date_attr(value),
        "min": _as_date_attr(min),
        "max": _as_date_attr(max),
    }
    session.send_input_message(id, drop_none(msg))


@add_example()
@doc_format(note=_note)
def update_date_range(
    id: str,
    *,
    label: Optional[str] = None,
    start: Optional[date | str] = None,
    end: Optional[date | str] = None,
    min: Optional[date | str] = None,
    max: Optional[date | str] = None,
    session: Optional[Session] = None,
) -> None:
    """
    Change the start and end values of a date range input on the client.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    start
        The initial start date. Either a :class:`~datetime.date` object, or a string in
        yyyy-mm-dd format. If ``None`` (the default), will use the current date in the
        client's time zone.
    end
        The initial end date. Either a :class:`~datetime.date` object, or a string in
        yyyy-mm-dd format. If ``None`` (the default), will use the current date in the
        client's time zone.
    min
        The minimum allowed value.
    max
        The maximum allowed value.
    session
        A :class:`~shiny.Session` instance. If not provided, it is inferred via
        :func:`~shiny.session.get_current_session`.

    Note
    ----
    {note}

    See Also
    --------
    * :func:`~shiny.ui.input_date_range`
    """

    session = require_active_session(session)
    value = {"start": _as_date_attr(start), "end": _as_date_attr(end)}
    msg = {
        "label": label,
        "value": drop_none(value),
        "min": _as_date_attr(min),
        "max": _as_date_attr(max),
    }
    session.send_input_message(id, drop_none(msg))


# -----------------------------------------------------------------------------
# input_numeric.py
# -----------------------------------------------------------------------------
@add_example()
@doc_format(note=_note)
def update_numeric(
    id: str,
    *,
    label: Optional[str] = None,
    value: Optional[float] = None,
    min: Optional[float] = None,
    max: Optional[float] = None,
    step: Optional[float] = None,
    session: Optional[Session] = None,
) -> None:
    """
    Change the value of a number input on the client.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    value
        A new value.
    min
        The minimum allowed value.
    max
        The maximum allowed value.
    step
        Interval to use when stepping between min and max.
    session
        The :class:`~shiny.Session` object passed to the server function of a :class:`~shiny.App`.

    Note
    ----
    {note}

    See Also
    --------
    * :func:`~shiny.ui.input_numeric`
    """

    session = require_active_session(session)
    msg = {
        "label": label,
        "value": value,
        "min": min,
        "max": max,
        "step": step,
    }
    session.send_input_message(id, drop_none(msg))


# -----------------------------------------------------------------------------
# input_select.py
# -----------------------------------------------------------------------------
@add_example()
@doc_format(note=_note)
def update_select(
    id: str,
    *,
    label: Optional[str] = None,
    choices: Optional[SelectChoicesArg] = None,
    selected: Optional[str | list[str]] = None,
    session: Optional[Session] = None,
) -> None:
    """
    Change the value of a select input on the client.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    choices
        Either a list of choices or a dictionary mapping choice values to labels. Note
        that if a dictionary is provided, the keys are used as the (input) values so
        that the dictionary values can hold HTML labels. A dictionary of dictionaries is
        also supported, and in that case, the top-level keys are treated as
        ``<optgroup>`` labels.
    selected
        The values that should be initially selected, if any.
    session
        A :class:`~shiny.Session` instance. If not provided, it is inferred via
        :func:`~shiny.session.get_current_session`.

    Note
    ----
    {note}

    See Also
    --------
    * :func:`~shiny.ui.input_select`
    * :func:`~shiny.ui.update_selectize`
    """

    session = require_active_session(session)

    selected_values = selected
    if isinstance(selected, str):
        selected_values = [selected]

    if choices is None:
        options = None
    else:
        option_tags = _render_choices(_normalize_choices(choices), selected)
        # Typing problem due to a bug in pylance:
        # https://github.com/microsoft/pylance-release/issues/2377
        options = session._process_ui(option_tags)["html"]  # type: ignore

    msg = {
        "label": label,
        "options": options,
        "value": selected_values,
    }
    session.send_input_message(id, drop_none(msg))


class FlatSelectChoice(TypedDict):
    label: str
    value: str
    optgroup: NotRequired[str]


@add_example()
@doc_format(note=_note)
def update_selectize(
    id: str,
    *,
    label: Optional[str] = None,
    choices: Optional[SelectChoicesArg] = None,
    selected: Optional[str | list[str]] = None,
    options: Optional[dict[str, str | float | JSEval]] = None,
    server: bool = False,
    session: Optional[Session] = None,
) -> None:
    """
    Change the value of a selectize.js powered input on the client.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    choices
        Either a list of choices or a dictionary mapping choice values to labels. Note
        that if a dictionary is provided, the keys are used as the (input) values so
        that the dictionary values can hold HTML labels. A dictionary of dictionaries is
        also supported, and in that case, the top-level keys are treated as
        ``<optgroup>`` labels.
    selected
        The values that should be initially selected, if any.
    options
        Options to send to update, see `input_selectize` for details.
    server
        Whether to store choices on the server side, and load the select options
        dynamically on searching, instead of writing all choices into the page at once
        (i.e., only use the client-side version of selectize.js)
    session
        A :class:`~shiny.Session` instance. If not provided, it is inferred via
        :func:`~shiny.session.get_current_session`.

    Note
    ----
    {note}

    See Also
    --------
    * :func:`~shiny.ui.input_selectize`
    """

    session = require_active_session(session)

    if not server:
        return update_select(
            id, label=label, choices=choices, selected=selected, session=session
        )

    if options is not None:
        cfg = TagList(
            tags.script(
                json.dumps(options),
                type="application/json",
                data_for=id,
                data_eval=json.dumps(extract_js_keys(options)),
            )
        )
        session.send_input_message(id, drop_none({"config": cfg.get_html_string()}))

    # Transform choices to a list of dicts (this is the form the client wants)
    # [{"label": "Foo", "value": "foo", "optgroup": "foo"}, ...]
    flat_choices: list[FlatSelectChoice] = []
    if choices is not None:
        for k, v in _normalize_choices(choices).items():
            if not isinstance(v, Mapping):
                flat_choices.append(
                    FlatSelectChoice(value=k, label=session._process_ui(v)["html"])
                )
            else:  # The optgroup case
                flat_choices.extend(
                    [
                        FlatSelectChoice(
                            optgroup=k, value=k2, label=session._process_ui(v2)["html"]
                        )
                        for (k2, v2) in v.items()
                    ]
                )

    selected_values = selected
    if isinstance(selected, str):
        selected_values = [selected]

    # Find any selected choices now so we have them ready to send to the client
    if selected_values is None:
        selected_choices = []
    else:
        selected_choices = [x for x in flat_choices if x["value"] in selected_values]

    def selectize_choices_json(request: Request) -> Response:
        if choices is None:
            return Response([], status_code=200)

        # N.B. relevant query parameters that shiny.js setscan be found here
        # https://github.com/rstudio/shiny/blob/78d77ce/srcts/src/bindings/input/selectInput.ts#L138-L142
        qparams = request.query_params

        # The (space-separated) input value(s) in lower-case (for case-insensitive matching)
        keywords = set(re.split(r"\s+", qparams.get("query", "").lower()))

        # Also note that the user (at least someday) has the ability to customize any of
        # these options https://github.com/rstudio/shiny/blob/78d77ce/srcts/src/bindings/input/selectInput.ts#L231
        #
        # For most options this is fine, but searchField/valueField require some validation.

        # i.e. maxOptions (defaults to 1000)
        max_options = int(qparams.get("maxop", 1000))

        # i.e. searchConjunction (defaults to 'and', but can also be 'or')
        conjunction = any if qparams.get("conju", "and") == "or" else all

        # i.e. searchFields (defaults to ['label'])
        search_fields: list[str] = json.loads(qparams.get("field", "['label']"))
        if len(search_fields) == 0:
            raise ValueError("The selectize.js searchFields option must be non-empty")

        # For some odd (probably wrong) reason, shiny.js is wrapping searchFields in an additional array
        # https://github.com/rstudio/shiny/blob/78d77ce/srcts/src/bindings/input/selectInput.ts#L139
        # https://github.com/rstudio/shiny/blob/78d77c/R/update-input.R#L801
        if isinstance(search_fields[0], list):
            search_fields = search_fields[0]

        if set(search_fields).difference(set(["label", "value", "optgroup"])):
            raise ValueError(
                "The selectize.js searchFields option must contain some combination of: "
                + "'label', 'value', and 'optgroup'"
            )

        # i.e. valueField (defaults to 'value')
        if qparams.get("value", "value") != "value":
            raise ValueError(
                "The selectize.js valueField option must be set to 'value'"
            )

        filtered_choices: list[FlatSelectChoice] = []
        for choice in flat_choices:
            # Short-circuit if we've reached the max number of options
            if (len(filtered_choices) + len(selected_choices)) > max_options:
                break

            # If this is a selected value, *don't* add it here (add after this loop)
            if selected_values and choice["value"] in selected_values:
                continue

            match = False
            for f in search_fields:
                val: Optional[str] = choice.get(f, None)
                # optgroup could be requested, but not necessarily present/relevant
                if val is None:
                    continue
                if conjunction([x in val.lower() for x in keywords]):
                    match = True

            if match:
                filtered_choices.append(choice)

        if selected_choices:
            filtered_choices.extend(selected_choices)

        return JSONResponse(filtered_choices, status_code=200)

    msg = {
        "label": label,
        "value": selected_values,
        "url": session.dynamic_route(f"update_selectize_{id}", selectize_choices_json),
    }

    return session.send_input_message(id, drop_none(msg))


# -----------------------------------------------------------------------------
# input_slider.py
# -----------------------------------------------------------------------------
@add_example()
@doc_format(note=_note)
def update_slider(
    id: str,
    *,
    label: Optional[str] = None,
    value: Optional[SliderValueArg | tuple[SliderValueArg, SliderValueArg]] = None,
    min: Optional[SliderValueArg] = None,
    max: Optional[SliderValueArg] = None,
    step: Optional[SliderStepArg] = None,
    time_format: Optional[str] = None,
    timezone: Optional[str] = None,
    session: Optional[Session] = None,
) -> None:
    """
    Change the value of a slider input on the client.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    value
        A new value.
    min
        The minimum allowed value.
    max
        The maximum allowed value.
    step
        Specifies the interval between each selectable value on the slider. Either
        ``None`` (the default), which uses a heuristic to determine the step size or a
        single number. If the values are dates, step is in days; if the values are
        date-times, step is in seconds.
    time_format
        Only used if the slider values are :class:`~datetime.date` or
        :class:`~datetime.datetime` objects. A time format string, to be passed to the
        Javascript strftime library. See https://github.com/samsonjs/strftime for more
        details. For Dates, the default is "%F" (like "2015-07-01"), and for Datetimes,
        the default is "%F %T" (like "2015-07-01 15:32:10").
    timezone
        Only used if the values are :class:`~datetime.datetime` objects. A string
        specifying the time zone offset for the displayed times, in the format "+HHMM"
        or "-HHMM". If ``None`` (the default), times will be displayed in the browser's
        time zone. The value "+0000" will result in UTC time.
    session
        A :class:`~shiny.Session` instance. If not provided, it is inferred via
        :func:`~shiny.session.get_current_session`.

    Note
    ----
    {note}

    See Also
    --------
    * :func:`~shiny.ui.input_slider`
    """

    session = require_active_session(session)

    # Get any non-None value to see if the `data-type` may need to change
    val = value[0] if isinstance(value, (tuple, list)) else value
    present_val = next((x for x in [val, min, max] if x is not None), None)

    data_type = None if present_val is None else _slider_type(present_val)
    if time_format is None and data_type and data_type[0:4] == "date":
        time_format = "%F" if data_type == "date" else "%F %T"

    min_num = None if min is None else _as_numeric(min)
    max_num = None if max is None else _as_numeric(max)
    step_num = None if step is None else _as_numeric(step)
    if isinstance(value, (tuple, list)):
        value_num = [_as_numeric(x) for x in value]
    elif value is not None:
        value_num = _as_numeric(value)
    else:
        value_num = None

    msg = {
        "label": label,
        "value": value_num,
        "min": min_num,
        "max": max_num,
        "step": step_num,
        "data-type": data_type,
        "time_format": time_format,
        "timezone": timezone,
    }
    session.send_input_message(id, drop_none(msg))


# -----------------------------------------------------------------------------
# input_text.py
# -----------------------------------------------------------------------------
@add_example()
@doc_format(note=_note)
def update_text(
    id: str,
    *,
    label: Optional[str] = None,
    value: Optional[str] = None,
    placeholder: Optional[str] = None,
    session: Optional[Session] = None,
) -> None:
    """
    Change the value of a text input on the client.

    Parameters
    ----------
    id
        An input id.
    label
        An input label.
    value
        A new value.
    placeholder
        A hint as to what can be entered into the control.
    session
        A :class:`~shiny.Session` instance. If not provided, it is inferred via
        :func:`~shiny.session.get_current_session`.

    Note
    ----
    {note}

    See Also
    --------
    * :func:`~shiny.ui.input_text`
    """

    session = require_active_session(session)
    msg = {"label": label, "value": value, "placeholder": placeholder}
    session.send_input_message(id, drop_none(msg))


update_text_area = update_text
update_text_area.__doc__ = update_text.__doc__


# -----------------------------------------------------------------------------
# navs.py
# -----------------------------------------------------------------------------


# TODO: we should probably provide a nav_select() alias for this as well
@add_example()
@doc_format(note=_note)
def update_navs(
    id: str, selected: Optional[str] = None, session: Optional[Session] = None
) -> None:
    """
    Change the value of a navs container on the client.

    Parameters
    ----------
    id
        An input id.
    selected
        The values that should be initially selected, if any.
    session
        A :class:`~shiny.Session` instance. If not provided, it is inferred via
        :func:`~shiny.session.get_current_session`.

    Note
    ----
    {note}

    See Also
    --------
    * :func:`~shiny.ui.navset_tab`
    * :func:`~shiny.ui.navset_pill`
    * :func:`~shiny.ui.page_navbar`
    """

    session = require_active_session(session)
    msg = {"value": selected}
    session.send_input_message(id, drop_none(msg))


# -----------------------------------------------------------------------------
# tooltips.py
# -----------------------------------------------------------------------------
@add_example()
def update_tooltip(
    id: str,
    *args: TagChild,
    show: Optional[bool] = None,
    session: Optional[Session] = None,
) -> None:
    """
    Update tooltip contents.

    Parameters
    ----------
    id
        A character string that matches an existing tooltip id.
    *args
        Contents to the tooltip's body.
    show
        Opens (`True`) or closes (`False`) the tooltip.
    session
        A Shiny session object (the default should almost always be used).
    """

    _session_on_flush_send_msg(
        id,
        session,
        drop_none(
            {
                "method": "update",
                "title": (
                    require_active_session(session)._process_ui(TagList(*args))
                    if len(args) > 0
                    else None
                ),
            }
        ),
    )
    if show is not None:
        _session_on_flush_send_msg(
            id,
            session,
            {
                "method": "toggle",
                "value": _normalize_show_value(show),
            },
        )


# -----------------------------------------------------------------------------
# popover.py
# -----------------------------------------------------------------------------


@add_example()
def update_popover(
    id: str,
    *args: TagChild,
    title: Optional[TagChild] = None,
    show: Optional[bool] = None,
    session: Optional[Session] = None,
) -> None:
    """
    Update the contents or title of a popover.

    Parameters
    ----------
    id
        The id of the popover DOM element to update.
    *args
        The new contents of the popover.
    title
        The new title of the popover.
    show
        Opens (`True`) or closes (`False`) the popover.
    session
        A Shiny session object (the default should almost always be used).

    See Also
    --------
    * :func:`~shiny.ui.popover`
    """
    session = require_active_session(session)

    if title is not None or len(args) > 0:
        _session_on_flush_send_msg(
            id,
            session,
            drop_none(
                {
                    "method": "update",
                    "content": (
                        session._process_ui(TagList(*args)) if len(args) > 0 else None
                    ),
                    "header": session._process_ui(title) if title is not None else None,
                },
            ),
        )
    if show is not None:
        _session_on_flush_send_msg(
            id,
            session,
            {
                "method": "toggle",
                "value": _normalize_show_value(show),
            },
        )


@overload
def _normalize_show_value(show: None) -> Literal["toggle"]: ...


@overload
def _normalize_show_value(show: bool) -> Literal["show", "hide"]: ...


def _normalize_show_value(show: bool | None) -> Literal["toggle", "show", "hide"]:
    if show is None:
        return "toggle"
    return "show" if show else "hide"



from faicons import icon_svg

from shiny.express import input, render, ui

ui.page_opts(title="Kitchen Sink: ui.input_action_link()", fillable=True)

with ui.layout_columns():
    with ui.card():
        ui.input_action_link("default", label="Default action link")

        @render.code
        def default_txt():
            return f"Link clicked {input.default()} times"

    with ui.card():
        ui.input_action_link("icon", "Link with icon", icon=icon_svg("trash-arrow-up"))

        @render.code
        def icon_txt():
            return f"Link clicked {input.icon()} times"



import re

from playwright.sync_api import Page, expect
from utils.deploy_utils import skip_on_webkit

from shiny.playwright import controller
from shiny.run import ShinyAppProc


@skip_on_webkit
def test_validate_chat(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    chat = controller.Chat(page, "chat")
    message_state = controller.OutputCode(page, "message_state")

    # Wait for app to load
    message_state.expect.not_to_have_text("", timeout=30 * 1000)

    expect(chat.loc).to_be_visible()
    expect(chat.loc_input_button).to_be_disabled()

    messages = [
        "FIRST FIRST FIRST",
        "SECOND SECOND SECOND",
        "THIRD THIRD THIRD",
        "FOURTH FOURTH FOURTH",
        "FIFTH FIFTH FIFTH",
    ]
    # Allow for any whitespace between messages
    chat.expect_messages(re.compile(r"\s*".join(messages)))

    message_state_expected = tuple(
        [{"content": message, "role": "assistant"} for message in messages]
    )
    message_state.expect_value(str(message_state_expected))



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_input_text_area_kitchen(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    default = controller.InputTextArea(page, "default")
    default.expect_label("Default text area")
    default.expect_value("")
    controller.OutputCode(page, "default_txt").expect.to_have_text("")

    placeholder = controller.InputTextArea(page, "placeholder")
    placeholder.expect_placeholder("Enter text here")
    placeholder.expect_value("")

    custom_size = controller.InputTextArea(page, "custom_size")
    custom_size.expect_height("150px")
    custom_size.expect_width("300px")

    custom_cols = controller.InputTextArea(page, "cols")
    custom_cols.expect_cols("30")

    custom_rows = controller.InputTextArea(page, "rows")
    custom_rows.expect_rows("5")
    custom_rows.expect_resize("none")

    autocomplete = controller.InputTextArea(page, "autocomplete")
    autocomplete.expect_autocomplete("name")
    autocomplete.expect_resize("horizontal")

    resize = controller.InputTextArea(page, "resize")
    resize.expect_resize("both")
    resize.set("Some text")
    resize.expect_value("Some text")
    controller.OutputCode(page, "resize_txt").expect.to_have_text("Some text")

    spellcheck = controller.InputTextArea(page, "spellcheck")
    spellcheck.expect_spellcheck("true")
    spellcheck.expect_resize("vertical")



"""Tests for """

import contextlib
import random
import textwrap
from typing import Generator

from htmltools import TagList

from shiny import ui
from shiny._utils import private_seed


# Fix the randomness of these functions to make the tests deterministic
@contextlib.contextmanager
def private_seed_n(n: int = 0) -> Generator[None, None, None]:
    with private_seed():
        random.seed(n)
        yield


a = ui.nav_panel("a", "a")
b = ui.nav_panel("b", "b")
c = ui.nav_panel("c", "c")
menu = ui.nav_menu(
    "Menu",
    c,
    "----",
    "Plain text",
    "----",
    ui.nav_control("Other item"),
)


def test_navset_tab_markup():
    with private_seed_n():
        x = ui.navset_tab(a, b, ui.nav_control("Some item"), menu)

    assert TagList(x).render()["html"] == textwrap.dedent(
        """\
        <ul class="nav nav-tabs" data-tabsetid="886440">
          <li class="nav-item">
            <a data-bs-toggle="tab" data-toggle="tab" data-value="a" role="tab" class="nav-link active" href="#tab-886440-0">a</a>
          </li>
          <li class="nav-item">
            <a data-bs-toggle="tab" data-toggle="tab" data-value="b" role="tab" class="nav-link" href="#tab-886440-1">b</a>
          </li>
          <li>Some item</li>
          <li class="nav-item dropdown">
            <a class="nav-link dropdown-toggle " data-bs-toggle="dropdown" data-value="Menu" href="#" role="button">Menu</a>
            <ul class="dropdown-menu" data-tabsetid="404958">
              <li>
                <a data-bs-toggle="tab" data-toggle="tab" data-value="c" role="tab" class="dropdown-item" href="#tab-404958-0">c</a>
              </li>
              <li class="dropdown-divider"></li>
              <li class="dropdown-header">Plain text</li>
              <li class="dropdown-divider"></li>
              <li>Other item</li>
            </ul>
          </li>
        </ul>
        <div class="tab-content" data-tabsetid="886440">
          <div class="tab-pane active" role="tabpanel" data-value="a" id="tab-886440-0">a</div>
          <div class="tab-pane" role="tabpanel" data-value="b" id="tab-886440-1">b</div>
          <div class="tab-pane" role="tabpanel" data-value="c" id="tab-404958-0">c</div>
        </div>"""
    )


def test_navset_pill_markup():
    with private_seed_n():
        x = ui.navset_pill(menu, a, id="navset_pill_id")

    assert TagList(x).render()["html"] == textwrap.dedent(
        """\
        <ul class="nav nav-pills shiny-tab-input" id="navset_pill_id" data-tabsetid="886440">
          <li class="nav-item dropdown">
            <a class="nav-link dropdown-toggle active" data-bs-toggle="dropdown" data-value="Menu" href="#" role="button">Menu</a>
            <ul class="dropdown-menu" data-tabsetid="404958">
              <li>
                <a data-bs-toggle="tab" data-toggle="tab" data-value="c" role="tab" class="dropdown-item active" href="#tab-404958-0">c</a>
              </li>
              <li class="dropdown-divider"></li>
              <li class="dropdown-header">Plain text</li>
              <li class="dropdown-divider"></li>
              <li>Other item</li>
            </ul>
          </li>
          <li class="nav-item">
            <a data-bs-toggle="tab" data-toggle="tab" data-value="a" role="tab" class="nav-link" href="#tab-886440-1">a</a>
          </li>
        </ul>
        <div class="tab-content" data-tabsetid="886440">
          <div class="tab-pane active" role="tabpanel" data-value="c" id="tab-404958-0">c</div>
          <div class="tab-pane" role="tabpanel" data-value="a" id="tab-886440-1">a</div>
        </div>"""
    )


def test_navset_card_pill_markup():
    with private_seed_n():
        x = ui.navset_card_pill(
            a,
            ui.nav_menu("Menu", c),
            b,
            selected="c",
        )

    assert TagList(x).render()["html"] == textwrap.dedent(
        """\
        <div class="card bslib-card bslib-mb-spacing html-fill-item html-fill-container" data-bslib-card-init="">
          <div class="card-header">
            <ul class="nav nav-pills card-header-pills" data-tabsetid="886440">
              <li class="nav-item">
                <a data-bs-toggle="tab" data-toggle="tab" data-value="a" role="tab" class="nav-link" href="#tab-886440-0">a</a>
              </li>
              <li class="nav-item dropdown">
                <a class="nav-link dropdown-toggle active" data-bs-toggle="dropdown" data-value="Menu" href="#" role="button">Menu</a>
                <ul class="dropdown-menu" data-tabsetid="404958">
                  <li>
                    <a data-bs-toggle="tab" data-toggle="tab" data-value="c" role="tab" class="dropdown-item active" href="#tab-404958-0">c</a>
                  </li>
                </ul>
              </li>
              <li class="nav-item">
                <a data-bs-toggle="tab" data-toggle="tab" data-value="b" role="tab" class="nav-link" href="#tab-886440-2">b</a>
              </li>
            </ul>
          </div>
          <div class="card-body bslib-gap-spacing html-fill-item html-fill-container" style="margin-top:auto;margin-bottom:auto;flex:1 1 auto;">
            <div class="tab-content html-fill-item html-fill-container" data-tabsetid="886440">
              <div class="tab-pane html-fill-item html-fill-container bslib-gap-spacing" role="tabpanel" data-value="a" id="tab-886440-0" style="gap:0;padding:0;">a</div>
              <div class="tab-pane active html-fill-item html-fill-container bslib-gap-spacing" role="tabpanel" data-value="c" id="tab-404958-0" style="gap:0;padding:0;">c</div>
              <div class="tab-pane html-fill-item html-fill-container bslib-gap-spacing" role="tabpanel" data-value="b" id="tab-886440-2" style="gap:0;padding:0;">b</div>
            </div>
          </div>
          <script data-bslib-card-init="">window.bslib.Card.initializeAllCards();</script>
        </div>"""
    )


def test_navset_bar_markup():
    with private_seed_n():
        x = ui.navset_bar(
            ui.nav_menu("Menu", "Plain text", c),
            title="Page title",
            footer="Page footer",
            header="Page header",
        )

    assert TagList(x).render()["html"] == textwrap.dedent(
        """\
        <nav class="navbar navbar-expand-md navbar-default">
          <div class="container-fluid">
            <span class="navbar-brand">Page title</span><button class="navbar-toggler" type="button" data-bs-toggle="collapse" data-bs-target="#navbar-collapse-795772" aria-controls="navbar-collapse-795772" aria-expanded="false" aria-label="Toggle navigation"><span class="navbar-toggler-icon"></span></button>
            <div id="navbar-collapse-795772" class="collapse navbar-collapse">
              <ul class="nav navbar-nav nav-underline" data-tabsetid="886440">
                <li class="nav-item dropdown">
                  <a class="nav-link dropdown-toggle active" data-bs-toggle="dropdown" data-value="Menu" href="#" role="button">Menu</a>
                  <ul class="dropdown-menu" data-tabsetid="404958">
                    <li class="dropdown-header">Plain text</li>
                    <li>
                      <a data-bs-toggle="tab" data-toggle="tab" data-value="c" role="tab" class="dropdown-item active" href="#tab-404958-1">c</a>
                    </li>
                  </ul>
                </li>
              </ul>
            </div>
          </div>
        </nav>
        <div class="container-fluid html-fill-item html-fill-container">
          Page header
          <div class="tab-content html-fill-item html-fill-container" data-tabsetid="886440">
            <div class="tab-pane active html-fill-item html-fill-container bslib-gap-spacing" role="tabpanel" data-value="c" id="tab-404958-1" style="--bslib-navbar-margin:0;;">c</div>
          </div>
          Page footer
        </div>"""
    )



from icons import piggy_bank

from shiny import App, ui

app_ui = ui.page_fluid(
    ui.layout_column_wrap(
        ui.value_box(
            "KPI Title",
            "$1 Billion Dollars",
            "Up 30% VS PREVIOUS 30 DAYS",
            showcase=piggy_bank,
            theme="bg-gradient-orange-red",
            full_screen=True,
        ),
        ui.value_box(
            "KPI Title",
            "$1 Billion Dollars",
            "Up 30% VS PREVIOUS 30 DAYS",
            showcase=piggy_bank,
            theme="text-green",
            showcase_layout="top right",
            full_screen=True,
        ),
        ui.value_box(
            "KPI Title",
            "$1 Billion Dollars",
            "Up 30% VS PREVIOUS 30 DAYS",
            showcase=piggy_bank,
            theme="purple",
            showcase_layout="bottom",
            full_screen=True,
        ),
    )
)


app = App(app_ui, server=None)



from __future__ import annotations

import collections.abc
import copy
import re
from typing import Any, Literal, Optional, Sequence, cast

from htmltools import (
    HTML,
    MetadataNode,
    Tag,
    TagAttrs,
    TagChild,
    TagList,
    css,
    div,
    tags,
)

from .._docstring import add_example
from .._namespaces import resolve_id_or_none
from .._utils import private_random_int
from ..types import NavSetArg
from ._bootstrap import column, row
from ._card import CardItem, WrapperCallable, card, card_body, card_footer, card_header
from ._html_deps_shinyverse import components_dependencies
from ._sidebar import Sidebar, layout_sidebar
from .css import CssUnit, as_css_padding, as_css_unit
from .fill import as_fill_item, as_fillable_container

__all__ = (
    "nav_panel",
    "nav_menu",
    "nav_control",
    "nav_spacer",
    "navset_tab",
    "navset_card_tab",
    "navset_pill",
    "navset_underline",
    "navset_card_pill",
    "navset_card_underline",
    "navset_pill_list",
    "navset_hidden",
    "navset_bar",
)


# -----------------------------------------------------------------------------
# Navigation items
# -----------------------------------------------------------------------------
class NavPanel:
    nav: Tag
    content: Optional[Tag]

    def __init__(self, nav: Tag, content: Optional[Tag] = None) -> None:
        self.nav = nav
        # nav_control()/nav_spacer() have None as their content
        self.content = content

    def resolve(
        self, selected: Optional[str], context: dict[str, Any]
    ) -> tuple[TagChild, TagChild]:
        # Nothing to do for nav_control()/nav_spacer()
        if self.content is None:
            return self.nav, None

        # At least currently, in the case where both nav and content are tags
        # (i.e., nav()), the nav always has a child <a> tag...I'm not sure if
        # there's a way to statically type this
        nav = copy.deepcopy(self.nav)
        a_tag = cast(Tag, nav.children[0])
        if context.get("is_menu", False):
            a_tag.add_class("dropdown-item")
        else:
            a_tag.add_class("nav-link")
            nav.add_class("nav-item")

        # Hyperlink the nav to the content
        content = copy.copy(self.content)
        if "tabsetid" in context and "index" in context:
            id = f"tab-{context['tabsetid']}-{context['index']}"
            content.attrs["id"] = id
            a_tag.attrs["href"] = f"#{id}"

        # Mark the nav/content as active if it should be
        if isinstance(selected, str) and selected == self.get_value():
            content.add_class("active")
            a_tag.add_class("active")

        nav.children[0] = a_tag

        return nav, content

    def get_value(self) -> str | None:
        if self.content is None:
            return None
        a_tag = cast(Tag, self.nav.children[0])
        data_value_attr = a_tag.attrs.get("data-value", None)
        if isinstance(data_value_attr, HTML):
            data_value_attr = str(data_value_attr)
        return data_value_attr

    def tagify(self) -> None:
        raise NotImplementedError(
            "nav() items must appear within navset_*() container."
        )


@add_example()
def nav_panel(
    title: TagChild,
    *args: TagChild,
    value: Optional[str] = None,
    icon: TagChild = None,
) -> NavPanel:
    """
    Create a nav item pointing to some internal content.

    Parameters
    ----------
    title
        A title to display. Can be a character string or UI elements (i.e., tags).
    *args
        UI elements to display when the item is active.
    value
        The value of the item. Use this value to determine whether the item is active
        (when an ``id`` is provided to the nav container) or to programmatically
        select the item (e.g., :func:`~shiny.ui.update_navs`). You can also
        provide the value to the ``selected`` argument of the navigation container
        (e.g., :func:`~shiny.ui.navset_tab`).
    icon
        An icon to appear inline with the button/link.

    See Also
    --------
    * :func:`~shiny.ui.nav_menu`
    * :func:`~shiny.ui.nav_control`
    * :func:`~shiny.ui.nav_spacer`
    * :func:`~shiny.ui.navset_bar`
    * :func:`~shiny.ui.navset_tab`
    * :func:`~shiny.ui.navset_pill`
    * :func:`~shiny.ui.navset_underline`
    * :func:`~shiny.ui.navset_card_tab`
    * :func:`~shiny.ui.navset_card_pill`
    * :func:`~shiny.ui.navset_card_underline`
    * :func:`~shiny.ui.navset_pill_list`
    * :func:`~shiny.ui.navset_hidden`
    """
    if value is None:
        value = str(title)

    # N.B. at this point, we don't have enough info to link the nav to the content
    # or add relevant classes. That's done later by consumers (i.e. nav containers)
    link = tags.a(
        icon,
        title,
        data_bs_toggle="tab",  # Bootstrap 5
        data_toggle="tab",  # Needed for shiny.js' insert-tab handler
        data_value=value,
        role="tab",
    )

    return NavPanel(
        tags.li(link),
        div(*args, class_="tab-pane", role="tabpanel", data_value=value),
    )


@add_example()
def nav_control(*args: TagChild) -> NavPanel:
    """
    Place a control in the navigation container.

    Parameters
    ----------
    *args
        UI elements to display as the nav item.

    See Also
    --------
    * :func:`~shiny.ui.nav_panel`
    * :func:`~shiny.ui.nav_menu`
    * :func:`~shiny.ui.nav_spacer`
    * :func:`~shiny.ui.navset_bar`
    * :func:`~shiny.ui.navset_tab`
    * :func:`~shiny.ui.navset_pill`
    * :func:`~shiny.ui.navset_underline`
    * :func:`~shiny.ui.navset_card_tab`
    * :func:`~shiny.ui.navset_card_pill`
    * :func:`~shiny.ui.navset_card_underline`
    * :func:`~shiny.ui.navset_pill_list`
    * :func:`~shiny.ui.navset_hidden`

    Example
    -------
    See :func:`~shiny.ui.nav_panel`
    """
    return NavPanel(tags.li(*args))


@add_example()
def nav_spacer() -> NavPanel:
    """
    Create space between nav items.

    See Also
    --------
    * :func:`~shiny.ui.nav_panel`
    * :func:`~shiny.ui.nav_menu`
    * :func:`~shiny.ui.nav_control`
    * :func:`~shiny.ui.navset_bar`
    * :func:`~shiny.ui.navset_tab`
    * :func:`~shiny.ui.navset_pill`
    * :func:`~shiny.ui.navset_underline`
    * :func:`~shiny.ui.navset_card_tab`
    * :func:`~shiny.ui.navset_card_pill`
    * :func:`~shiny.ui.navset_card_underline`
    * :func:`~shiny.ui.navset_pill_list`
    * :func:`~shiny.ui.navset_hidden`

    Example
    -------
    See :func:`~shiny.ui.nav_panel`
    """

    return NavPanel(tags.li(components_dependencies(), class_="bslib-nav-spacer"))


class NavMenu:
    nav_controls: list[NavSetArg]
    title: TagChild
    value: str
    align: Literal["left", "right"]

    def __init__(
        self,
        *args: NavSetArg | str,
        title: TagChild,
        value: str,
        align: Literal["left", "right"] = "left",
    ) -> None:
        self.nav_controls = [menu_string_as_nav(x) for x in args]
        self.title = title
        self.value = value
        self.align = align

    def resolve(
        self,
        selected: Optional[str],
        context: dict[str, Any],
    ) -> tuple[TagChild, TagChild]:
        nav, content = render_navset(
            *self.nav_controls,
            # `.dropdown-menu-right` is for Bootstrap 3; BS5 uses `.dropdown-menu-end`
            ul_class=f"dropdown-menu{' dropdown-menu-end' if self.align == 'right' else ''}",
            id=None,
            selected=selected,
            context={**context, "is_menu": True},
        )

        active = False
        for tab in content.children:
            if isinstance(tab, Tag) and tab.has_class("active"):
                active = True
                break

        return (
            tags.li(
                tags.a(
                    self.title,
                    class_=f"nav-link dropdown-toggle {'active' if active else ''}",
                    data_bs_toggle="dropdown",
                    # N.B. this value is only relevant for locating the insertion/removal
                    # of items inside the nav container
                    data_value=self.value,
                    href="#",
                    role="button",
                ),
                nav,
                class_="nav-item dropdown",
            ),
            content.children,
        )

    def get_value(self) -> str | None:
        for x in self.nav_controls:
            val = x.get_value()
            if val:
                return val
        return None

    def tagify(self) -> None:
        raise NotImplementedError("nav_menu() must appear within navset_*() container.")


def menu_string_as_nav(x: str | NavSetArg) -> NavSetArg:
    if not isinstance(x, str):
        return x

    if re.match(r"^\-+$", x):
        nav = tags.li(class_="dropdown-divider")
    else:
        nav = tags.li(x, class_="dropdown-header")

    return NavPanel(nav)


@add_example()
def nav_menu(
    title: TagChild,
    *args: NavPanel | str,
    value: Optional[str] = None,
    icon: TagChild = None,
    align: Literal["left", "right"] = "left",
) -> NavMenu:
    """
    Create a menu of nav items.

    Parameters
    ----------
    title
        A title to display. Can be a character string or UI elements (i.e., tags).
    *args
        A collection of nav items (e.g., :func:`~shiny.ui.nav_panel`) and/or strings.
        Strings will be rendered as a section header unless the string is a set
        of two or more hyphens (e.g., ``---``), in which case it will be rendered
        as a divider.
    value
        The value of the item. Use this value to determine whether the item is active
        (when an ``id`` is provided to the nav container) or to programmatically
        select the item (e.g., :func:`~shiny.ui.update_navs`). You can also
        provide the value to the ``selected`` argument of the navigation container
        (e.g., :func:`~shiny.ui.navset_tab`).
    icon
        An icon to appear inline with the button/link.
    align
        Horizontal alignment of the dropdown menu relative to dropdown toggle.

    Returns
    -------
    :
        A UI element representing both the navigation menu.

    See Also
    --------
    * :func:`~shiny.ui.nav_panel`
    * :func:`~shiny.ui.nav_control`
    * :func:`~shiny.ui.nav_spacer`
    * :func:`~shiny.ui.navset_bar`
    * :func:`~shiny.ui.navset_tab`
    * :func:`~shiny.ui.navset_pill`
    * :func:`~shiny.ui.navset_underline`
    * :func:`~shiny.ui.navset_card_tab`
    * :func:`~shiny.ui.navset_card_pill`
    * :func:`~shiny.ui.navset_card_underline`
    * :func:`~shiny.ui.navset_pill_list`
    * :func:`~shiny.ui.navset_hidden`

    Example
    -------
    See :func:`~shiny.ui.nav_panel`
    """
    if value is None:
        value = str(title)

    return NavMenu(
        *args,
        title=TagList(icon, title),
        value=value,
        align=align,
    )


class NavSet:
    args: tuple[NavSetArg | MetadataNode | Sequence[MetadataNode], ...]
    ul_class: str
    id: Optional[str]
    selected: Optional[str]
    header: TagChild
    footer: TagChild

    def __init__(
        self,
        *args: NavSetArg | MetadataNode | Sequence[MetadataNode],
        ul_class: str,
        id: Optional[str],
        selected: Optional[str],
        header: TagChild = None,
        footer: TagChild = None,
    ) -> None:
        self.args = args
        self.ul_class = ul_class
        self.id = id
        self.selected = selected
        self.header = header
        self.footer = footer

    def tagify(self) -> TagList | Tag:
        id = self.id
        ul_class = self.ul_class
        if id is not None:
            ul_class += " shiny-tab-input"

        nav, content = render_navset(
            *self.args, ul_class=ul_class, id=id, selected=self.selected, context={}
        )
        return self.layout(nav, content).tagify()

    def layout(self, nav: Tag, content: Tag) -> TagList | Tag:
        return TagList(nav, self.header, content, self.footer)


# -----------------------------------------------------------------------------
# Navigation containers
# -----------------------------------------------------------------------------
@add_example()
def navset_tab(
    *args: NavSetArg | MetadataNode | Sequence[MetadataNode],
    id: Optional[str] = None,
    selected: Optional[str] = None,
    header: TagChild = None,
    footer: TagChild = None,
) -> NavSet:
    """
    Render nav items as a tabset.

    Parameters
    ----------
    *args
        A collection of nav items (e.g., :func:`shiny.ui.nav_panel`).
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match it's
        ``value``).
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.


    See Also
    --------
    * :func:`~shiny.ui.nav_panel`
    * :func:`~shiny.ui.nav_menu`
    * :func:`~shiny.ui.nav_control`
    * :func:`~shiny.ui.nav_spacer`
    * :func:`~shiny.ui.navset_bar`
    * :func:`~shiny.ui.navset_pill`
    * :func:`~shiny.ui.navset_underline`
    * :func:`~shiny.ui.navset_card_tab`
    * :func:`~shiny.ui.navset_card_pill`
    * :func:`~shiny.ui.navset_card_underline`
    * :func:`~shiny.ui.navset_pill_list`
    * :func:`~shiny.ui.navset_hidden`

    Example
    -------
    See :func:`~shiny.ui.nav_panel`
    """

    return NavSet(
        *args,
        ul_class="nav nav-tabs",
        id=resolve_id_or_none(id),
        selected=selected,
        header=header,
        footer=footer,
    )


@add_example()
def navset_pill(
    *args: NavSetArg | MetadataNode | Sequence[MetadataNode],
    id: Optional[str] = None,
    selected: Optional[str] = None,
    header: TagChild = None,
    footer: TagChild = None,
) -> NavSet:
    """
    Render nav items as a pillset.

    Parameters
    ----------
    *args
        A collection of nav items (e.g., :func:`shiny.ui.nav_panel`).
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match it's
        ``value``).
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.

    See Also
    --------
    * :func:`~shiny.ui.nav_panel`
    * :func:`~shiny.ui.nav_menu`
    * :func:`~shiny.ui.nav_control`
    * :func:`~shiny.ui.nav_spacer`
    * :func:`~shiny.ui.navset_bar`
    * :func:`~shiny.ui.navset_tab`
    * :func:`~shiny.ui.navset_underline`
    * :func:`~shiny.ui.navset_card_tab`
    * :func:`~shiny.ui.navset_card_pill`
    * :func:`~shiny.ui.navset_card_underline`
    * :func:`~shiny.ui.navset_pill_list`
    * :func:`~shiny.ui.navset_hidden`

    Example
    -------
    See :func:`~shiny.ui.nav_panel`
    """

    return NavSet(
        *args,
        ul_class="nav nav-pills",
        id=resolve_id_or_none(id),
        selected=selected,
        header=header,
        footer=footer,
    )


@add_example()
def navset_underline(
    *args: NavSetArg | MetadataNode | Sequence[MetadataNode],
    id: Optional[str] = None,
    selected: Optional[str] = None,
    header: TagChild = None,
    footer: TagChild = None,
) -> NavSet:
    """
    Render nav items whose active/focused navigation links are styled with an underline.

    Parameters
    ----------
    *args
        A collection of nav items (e.g., :func:`shiny.ui.nav_panel`).
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match it's
        ``value``).
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.

    See Also
    --------
    * :func:`~shiny.ui.nav_panel`
    * :func:`~shiny.ui.nav_menu`
    * :func:`~shiny.ui.nav_control`
    * :func:`~shiny.ui.nav_spacer`
    * :func:`~shiny.ui.navset_bar`
    * :func:`~shiny.ui.navset_tab`
    * :func:`~shiny.ui.navset_pill`
    * :func:`~shiny.ui.navset_card_tab`
    * :func:`~shiny.ui.navset_card_pill`
    * :func:`~shiny.ui.navset_card_underline`
    * :func:`~shiny.ui.navset_pill_list`
    * :func:`~shiny.ui.navset_hidden`

    Example
    -------
    See :func:`~shiny.ui.nav_panel`
    """
    return NavSet(
        *args,
        ul_class="nav nav-underline",
        id=resolve_id_or_none(id),
        selected=selected,
        header=header,
        footer=footer,
    )


@add_example()
def navset_hidden(
    *args: NavSetArg | MetadataNode | Sequence[MetadataNode],
    id: Optional[str] = None,
    selected: Optional[str] = None,
    header: TagChild = None,
    footer: TagChild = None,
) -> NavSet:
    """
    Render nav contents without the nav items.

    Parameters
    ----------
    *args
        A collection of nav items (e.g., :func:`shiny.ui.nav_panel`).
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match it's
        ``value``).
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.

    See Also
    --------
    * :func:`~shiny.ui.nav_panel`
    * :func:`~shiny.ui.nav_menu`
    * :func:`~shiny.ui.nav_control`
    * :func:`~shiny.ui.nav_spacer`
    * :func:`~shiny.ui.navset_bar`
    * :func:`~shiny.ui.navset_tab`
    * :func:`~shiny.ui.navset_pill`
    * :func:`~shiny.ui.navset_underline`
    * :func:`~shiny.ui.navset_card_tab`
    * :func:`~shiny.ui.navset_card_pill`
    * :func:`~shiny.ui.navset_card_underline`
    * :func:`~shiny.ui.navset_pill_list`
    """

    return NavSet(
        *args,
        ul_class="nav nav-hidden",
        id=resolve_id_or_none(id),
        selected=selected,
        header=header,
        footer=footer,
    )


class NavSetCard(NavSet):
    placement: Literal["above", "below"]
    sidebar: Optional[Sidebar]
    title: Optional[TagChild]

    def __init__(
        self,
        *args: NavSetArg | MetadataNode | Sequence[MetadataNode],
        ul_class: str,
        id: Optional[str],
        selected: Optional[str],
        title: Optional[TagChild] = None,
        sidebar: Optional[Sidebar] = None,
        header: TagChild = None,
        footer: TagChild = None,
        placement: Literal["above", "below"] = "above",
    ) -> None:
        super().__init__(
            *args,
            ul_class=ul_class,
            id=id,
            selected=selected,
            header=header,
            footer=footer,
        )
        self.title = title
        self.sidebar = sidebar
        self.placement = placement

    def layout(self, nav: Tag, content: Tag) -> Tag:
        content = _make_tabs_fillable(content, fillable=True, gap=0, padding=0)

        contents: list[CardItem] = wrap_each_content(
            [
                child
                for child in [self.header, content, self.footer]
                if child is not None
            ]
        )

        # If there is a sidebar, make a size 1 array of the layout_sidebar content
        if self.sidebar:
            contents = [
                layout_sidebar(
                    self.sidebar,
                    *contents,
                    fillable=True,
                    border=False,
                )
            ]

        nav_items = [*navset_title(self.title), nav]

        return card(
            card_header(*nav_items) if self.placement == "above" else None,
            *contents,
            card_footer(*nav_items) if self.placement == "below" else None,
        )


@add_example()
def navset_card_tab(
    *args: NavSetArg | MetadataNode | Sequence[MetadataNode],
    id: Optional[str] = None,
    selected: Optional[str] = None,
    title: Optional[TagChild] = None,
    sidebar: Optional[Sidebar] = None,
    header: TagChild = None,
    footer: TagChild = None,
) -> NavSetCard:
    """
    Render nav items as a tabset inside a card container.

    Parameters
    ----------
    *args
        A collection of nav items (e.g., :func:`shiny.ui.nav_panel`).
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match it's
        ``value``).
    sidebar
        A `Sidebar` component to display on every `nav()` page.
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.

    See Also
    --------
    * :func:`~shiny.ui.nav_panel`
    * :func:`~shiny.ui.nav_menu`
    * :func:`~shiny.ui.nav_control`
    * :func:`~shiny.ui.nav_spacer`
    * :func:`~shiny.ui.navset_bar`
    * :func:`~shiny.ui.navset_tab`
    * :func:`~shiny.ui.navset_pill`
    * :func:`~shiny.ui.navset_underline`
    * :func:`~shiny.ui.navset_card_pill`
    * :func:`~shiny.ui.navset_card_underline`
    * :func:`~shiny.ui.navset_pill_list`
    * :func:`~shiny.ui.navset_hidden`

    Example
    -------
    See :func:`~shiny.ui.nav_panel`
    """

    return NavSetCard(
        *args,
        ul_class="nav nav-tabs card-header-tabs",
        id=resolve_id_or_none(id),
        selected=selected,
        title=title,
        sidebar=sidebar,
        header=header,
        footer=footer,
        placement="above",
    )


@add_example()
def navset_card_pill(
    *args: NavSetArg | MetadataNode | Sequence[MetadataNode],
    id: Optional[str] = None,
    selected: Optional[str] = None,
    title: Optional[TagChild] = None,
    sidebar: Optional[Sidebar] = None,
    header: TagChild = None,
    footer: TagChild = None,
    placement: Literal["above", "below"] = "above",
) -> NavSetCard:
    """
    Render nav items as a pillset inside a card container.

    Parameters
    ----------
    *args
        A collection of nav items (e.g., :func:`shiny.ui.nav_panel`).
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match it's
        ``value``).
    sidebar
        A :class:`shiny.ui.Sidebar` component to display on every :func:`~shiny.ui.nav_panel` page.
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.
    placement
        Placement of the nav items relative to the content.

    See Also
    --------
    * :func:`~shiny.ui.nav_panel`
    * :func:`~shiny.ui.nav_menu`
    * :func:`~shiny.ui.nav_control`
    * :func:`~shiny.ui.nav_spacer`
    * :func:`~shiny.ui.navset_bar`
    * :func:`~shiny.ui.navset_tab`
    * :func:`~shiny.ui.navset_pill`
    * :func:`~shiny.ui.navset_underline`
    * :func:`~shiny.ui.navset_card_tab`
    * :func:`~shiny.ui.navset_card_underline`
    * :func:`~shiny.ui.navset_pill_list`
    * :func:`~shiny.ui.navset_hidden`

    Example
    -------
    See :func:`~shiny.ui.nav_panel`
    """

    return NavSetCard(
        *args,
        ul_class="nav nav-pills card-header-pills",
        id=resolve_id_or_none(id),
        selected=selected,
        title=title,
        sidebar=sidebar,
        header=header,
        footer=footer,
        placement=placement,
    )


@add_example()
def navset_card_underline(
    *args: NavSetArg | MetadataNode | Sequence[MetadataNode],
    id: Optional[str] = None,
    selected: Optional[str] = None,
    title: Optional[TagChild] = None,
    sidebar: Optional[Sidebar] = None,
    header: TagChild = None,
    footer: TagChild = None,
    placement: Literal["above", "below"] = "above",
) -> NavSetCard:
    """
    Render nav items inside a card container. Active/focused navigation links are styled with an underline.

    Parameters
    ----------
    *args
        A collection of nav items (e.g., :func:`shiny.ui.nav_panel`).
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match it's
        ``value``).
    sidebar
        A :class:`shiny.ui.Sidebar` component to display on every :func:`~shiny.ui.nav_panel` page.
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.
    placement
        Placement of the nav items relative to the content.

    See Also
    --------
    * :func:`~shiny.ui.nav_panel`
    * :func:`~shiny.ui.nav_menu`
    * :func:`~shiny.ui.nav_control`
    * :func:`~shiny.ui.nav_spacer`
    * :func:`~shiny.ui.navset_bar`
    * :func:`~shiny.ui.navset_tab`
    * :func:`~shiny.ui.navset_pill`
    * :func:`~shiny.ui.navset_underline`
    * :func:`~shiny.ui.navset_card_tab`
    * :func:`~shiny.ui.navset_card_pill`
    * :func:`~shiny.ui.navset_pill_list`
    * :func:`~shiny.ui.navset_hidden`

    Example
    -------
    See :func:`~shiny.ui.nav_panel`
    """
    return NavSetCard(
        *args,
        ul_class="nav nav-underline",
        id=resolve_id_or_none(id),
        selected=selected,
        title=title,
        sidebar=sidebar,
        header=header,
        footer=footer,
        placement=placement,
    )


class NavSetPillList(NavSet):
    well: bool
    widths: tuple[int, int]

    def __init__(
        self,
        *args: NavSetArg | MetadataNode | Sequence[MetadataNode],
        ul_class: str,
        id: Optional[str],
        selected: Optional[str],
        header: TagChild = None,
        footer: TagChild = None,
        well: bool = True,
        widths: tuple[int, int] = (4, 8),
    ) -> None:
        super().__init__(
            *args,
            ul_class=ul_class,
            id=id,
            selected=selected,
            header=header,
            footer=footer,
        )
        self.well = well
        self.widths = widths

    def layout(self, nav: TagChild, content: TagChild) -> Tag:
        widths = self.widths
        return row(
            column(widths[0], nav, class_="well" if self.well else None),
            column(widths[1], self.header, content, self.footer),
        )


@add_example()
def navset_pill_list(
    *args: NavSetArg | MetadataNode | Sequence[MetadataNode],
    id: Optional[str] = None,
    selected: Optional[str] = None,
    header: TagChild = None,
    footer: TagChild = None,
    well: bool = True,
    widths: tuple[int, int] = (4, 8),
) -> NavSet:
    """
    Render nav items as a vertical pillset.

    Parameters
    ----------
    *args
        A collection of nav items (e.g., :func:`shiny.ui.nav_panel`).
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match its
        ``value``).
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.
    well
        ``True`` to place a well (gray rounded rectangle) around the navigation list.
    widths
        Column widths of the navigation list and tabset content areas respectively.

    See Also
    --------
    * :func:`~shiny.ui.nav_panel`
    * :func:`~shiny.ui.nav_menu`
    * :func:`~shiny.ui.nav_control`
    * :func:`~shiny.ui.nav_spacer`
    * :func:`~shiny.ui.navset_bar`
    * :func:`~shiny.ui.navset_tab`
    * :func:`~shiny.ui.navset_pill`
    * :func:`~shiny.ui.navset_underline`
    * :func:`~shiny.ui.navset_card_tab`
    * :func:`~shiny.ui.navset_card_pill`
    * :func:`~shiny.ui.navset_card_underline`
    * :func:`~shiny.ui.navset_hidden`

    Example
    -------
    See :func:`~shiny.ui.nav_panel`
    """

    return NavSetPillList(
        *args,
        ul_class="nav nav-pills nav-stacked",
        id=resolve_id_or_none(id),
        selected=selected,
        header=header,
        footer=footer,
        well=well,
        widths=widths,
    )


class NavSetBar(NavSet):
    title: TagChild
    sidebar: Optional[Sidebar]
    fillable: bool | list[str]
    gap: Optional[CssUnit]
    padding: Optional[CssUnit | list[CssUnit]]
    position: Literal["static-top", "fixed-top", "fixed-bottom", "sticky-top"]
    bg: Optional[str]
    inverse: bool
    underline: bool
    collapsible: bool
    fluid: bool
    _is_page_level: bool

    def __init__(
        self,
        *args: NavSetArg | MetadataNode | Sequence[MetadataNode],
        ul_class: str,
        title: TagChild,
        id: Optional[str],
        selected: Optional[str],
        sidebar: Optional[Sidebar] = None,
        fillable: bool | list[str] = False,
        gap: Optional[CssUnit],
        padding: Optional[CssUnit | list[CssUnit]],
        position: Literal[
            "static-top", "fixed-top", "fixed-bottom", "sticky-top"
        ] = "static-top",
        header: TagChild = None,
        footer: TagChild = None,
        bg: Optional[str] = None,
        # TODO: default to 'auto', like we have in R (parse color via webcolors?)
        inverse: bool = False,
        underline: bool = True,
        collapsible: bool = True,
        fluid: bool = True,
    ) -> None:
        super().__init__(
            *args,
            ul_class=ul_class,
            id=id,
            selected=selected,
            header=header,
            footer=footer,
        )
        self.title = title
        self.sidebar = sidebar
        self.fillable = fillable
        self.gap = gap
        self.padding = padding
        self.position = position
        self.bg = bg
        self.inverse = inverse
        self.underline = underline
        self.collapsible = collapsible
        self.fluid = fluid
        self._is_page_level = False

    def layout(self, nav: Tag, content: Tag) -> TagList:
        nav_container = div(
            {"class": "container-fluid" if self.fluid else "container"},
            tags.span({"class": "navbar-brand"}, self.title),
        )
        if self.collapsible:
            collapse_id = "navbar-collapse-" + nav_random_int()
            nav_container.append(
                tags.button(
                    tags.span(class_="navbar-toggler-icon"),
                    class_="navbar-toggler",
                    type="button",
                    data_bs_toggle="collapse",
                    data_bs_target="#" + collapse_id,
                    aria_controls=collapse_id,
                    aria_expanded="false",
                    aria_label="Toggle navigation",
                )
            )
            nav = div(nav, id=collapse_id, class_="collapse navbar-collapse")

        nav_container.append(nav)
        nav_final = tags.nav({"class": "navbar navbar-expand-md"}, nav_container)

        if self.position != "static-top":
            nav_final.add_class(self.position)

        # bslib supports navbar-default/navbar-inverse (which is no longer
        # a thing in Bootstrap 5) in a way that's still useful, especially Bootswatch.
        nav_final.add_class(f"navbar-{'inverse' if self.inverse else 'default'}")

        if self.bg:
            nav_final.add_style(f"background-color: {self.bg} !important;")

        content = _make_tabs_fillable(
            content,
            self.fillable,
            gap=self.gap,
            padding=self.padding,
            navbar=True,
        )

        # 2023-05-11; Do not wrap `row()` around `self.header` and `self.footer`
        contents: list[TagChild] = [
            child for child in [self.header, content, self.footer] if child is not None
        ]

        if self.sidebar is None:
            content_div = div(
                *contents,
                class_="container-fluid" if self.fluid else "container",
            )
            # If fillable is truthy, the .container also needs to be fillable
            if self.fillable:
                content_div = as_fillable_container(as_fill_item(content_div))
        else:
            tab_content = contents
            if self._is_page_level:
                # TODO-future: This could also be applied to the non-sidebar page layout above
                from ._page import page_main_container

                tab_content = page_main_container(*contents)

            content_div = div(
                # In the fluid case, the sidebar layout should be flush (i.e.,
                # the .container-fluid class adds padding that we don't want)
                {"class": "container"} if not self.fluid else None,
                layout_sidebar(
                    self.sidebar,
                    tab_content,
                    fillable=self.fillable is not False,
                    border_radius=False,
                    border=not self.fluid,
                ),
            )
            # Always have the sidebar layout fill its parent (in this case
            # fillable controls whether the _main_ content portion is fillable)
            content_div = as_fillable_container(as_fill_item(content_div))

        return TagList(nav_final, content_div)


# Given a .tab-content container, mark each relevant .tab-pane as a fill container/item.
def _make_tabs_fillable(
    content: Tag,
    fillable: bool | list[str] = True,
    gap: Optional[CssUnit] = None,
    padding: Optional[CssUnit | list[CssUnit]] = None,
    navbar: bool = False,
) -> Tag:
    if not fillable:
        return content

    # Even if only one .tab-pane wants fillable behavior, the .tab-content
    # must to be a fillable container.
    content = as_fillable_container(as_fill_item(content))

    for i, child in enumerate(content.children):
        # Only work on Tags
        if not isinstance(child, Tag):
            continue
        # Only work on .tab-pane children
        if not child.has_class("tab-pane"):
            continue
        # If `fillable` is a list, only fill the .tab-pane if its data-value is contained in `fillable`
        if isinstance(fillable, list):
            child_attr = child.attrs.get("data-value")
            if child_attr is None or child_attr not in fillable:
                continue
        styles = css(
            gap=as_css_unit(gap),
            padding=as_css_padding(padding),
            __bslib_navbar_margin="0;" if navbar else None,
        )
        child = as_fillable_container(as_fill_item(child))
        child.add_style(cast(str, styles))
        child.add_class("bslib-gap-spacing")

        content.children[i] = child

    return content


# TODO-future; Content should not be indented unless when called from `page_navbar()`
@add_example()
def navset_bar(
    *args: NavSetArg | MetadataNode | Sequence[MetadataNode],
    title: TagChild,
    id: Optional[str] = None,
    selected: Optional[str] = None,
    sidebar: Optional[Sidebar] = None,
    fillable: bool | list[str] = True,
    gap: Optional[CssUnit] = None,
    padding: Optional[CssUnit | list[CssUnit]] = None,
    position: Literal[
        "static-top", "fixed-top", "fixed-bottom", "sticky-top"
    ] = "static-top",
    header: TagChild = None,
    footer: TagChild = None,
    bg: Optional[str] = None,
    # TODO: default to 'auto', like we have in R (parse color via webcolors?)
    inverse: bool = False,
    underline: bool = True,
    collapsible: bool = True,
    fluid: bool = True,
) -> NavSetBar:
    """
    Render nav items as a navbar.

    Parameters
    ----------
    *args
        A collection of nav items (e.g., :func:`shiny.ui.nav_panel`).
    title
        Title to display in the navbar.
    id
        If provided, will create an input value that holds the currently selected nav
        item.
    selected
        Choose a particular nav item to select by default value (should match it's
        ``value``).
    sidebar
        A :class:`~shiny.ui.Sidebar` component to display on every :func:`~shiny.ui.nav_panel` page.
    fillable
        Whether or not to allow fill items to grow/shrink to fit the browser window. If
        `True`, all `nav()` pages are fillable. A character vector, matching the value
        of `nav()`s to be filled, may also be provided. Note that, if a `sidebar` is
        provided, `fillable` makes the main content portion fillable.
    gap
        A CSS length unit defining the gap (i.e., spacing) between elements provided to
        `*args`. This value is only used when the navbar is `fillable`.
    padding
        Padding to use for the body. This can be a numeric vector (which will be
        interpreted as pixels) or a character vector with valid CSS lengths. The length
        can be between one and four. If one, then that value will be used for all four
        sides. If two, then the first value will be used for the top and bottom, while
        the second value will be used for left and right. If three, then the first will
        be used for top, the second will be left and right, and the third will be
        bottom. If four, then the values will be interpreted as top, right, bottom, and
        left respectively. This value is only used when the navbar is `fillable`.
    position
        Determines whether the navbar should be displayed at the top of the page with
        normal scrolling behavior ("static-top"), pinned at the top ("fixed-top"), or
        pinned at the bottom ("fixed-bottom"). Note that using "fixed-top" or
        "fixed-bottom" will cause the navbar to overlay your body content, unless you
        add padding (e.g., ``tags.style("body {padding-top: 70px;}")``).
    header
        UI to display above the selected content.
    footer
        UI to display below the selected content.
    bg
        Background color of the navbar (a CSS color).
    inverse
        Either ``True`` for a light text color or ``False`` for a dark text color.
    collapsible
        ``True`` to automatically collapse the navigation elements into an expandable menu on mobile devices or narrow window widths.
    fluid
        ``True`` to use fluid layout; ``False`` to use fixed layout.

    See Also
    --------
    * :func:`~shiny.ui.page_navbar`
    * :func:`~shiny.ui.nav_panel`
    * :func:`~shiny.ui.nav_menu`
    * :func:`~shiny.ui.nav_control`
    * :func:`~shiny.ui.nav_spacer`
    * :func:`~shiny.ui.navset_tab`
    * :func:`~shiny.ui.navset_pill`
    * :func:`~shiny.ui.navset_underline`
    * :func:`~shiny.ui.navset_card_tab`
    * :func:`~shiny.ui.navset_card_pill`
    * :func:`~shiny.ui.navset_card_underline`
    * :func:`~shiny.ui.navset_pill_list`
    * :func:`~shiny.ui.navset_hidden`

    Example
    -------
    See :func:`~shiny.ui.nav_panel`.
    """

    # If args contains any lists, flatten them into args.
    new_args: Sequence[NavSetArg | MetadataNode] = []
    for arg in args:
        if isinstance(arg, (list, tuple)):
            new_args.extend(arg)
        else:
            new_args.append(cast(NavSetArg, arg))

    ul_class = "nav navbar-nav"
    if underline:
        ul_class += " nav-underline"

    return NavSetBar(
        *new_args,
        ul_class=ul_class,
        id=resolve_id_or_none(id),
        selected=selected,
        sidebar=sidebar,
        fillable=fillable,
        gap=gap,
        padding=padding,
        title=title,
        position=position,
        header=header,
        footer=footer,
        bg=bg,
        inverse=inverse,
        underline=underline,
        collapsible=collapsible,
        fluid=fluid,
    )


# -----------------------------------------------------------------------------
# Utilities for rendering navs
# -----------------------------------------------------------------------------\
def render_navset(
    *items: NavSetArg | MetadataNode | Sequence[MetadataNode],
    ul_class: str,
    id: Optional[str],
    selected: Optional[str],
    context: dict[str, Any],
) -> tuple[Tag, Tag]:
    tabsetid = nav_random_int()

    # Separate MetadataNodes from NavSetArgs.
    metadata_args: list[MetadataNode] = []
    navset_args: list[NavSetArg] = []

    for item in items:
        if isinstance(item, MetadataNode):
            metadata_args.append(item)
        elif isinstance(item, collections.abc.Sequence) and all(
            isinstance(x, MetadataNode) for x in item
        ):
            # Above we needed to use collections.abc.Sequence for runtime checks, as
            # typing.Sequence does not work for runtime checks.
            metadata_args.extend(item)
        else:
            # pyright needs a little help with type inference here.
            navset_args.append(cast(NavSetArg, item))

    # If the user hasn't provided a selected value, use the first one
    if selected is None:
        for x in navset_args:
            selected = x.get_value()
            if selected is not None:
                break

    ul_tag = tags.ul(
        metadata_args,
        class_=ul_class,
        id=id,
        data_tabsetid=tabsetid,
    )
    div_tag = div(class_="tab-content", data_tabsetid=tabsetid)
    for i, x in enumerate(navset_args):
        nav, contents = x.resolve(
            selected, {**context, "tabsetid": tabsetid, "index": i}
        )
        ul_tag.append(nav)
        div_tag.append(contents)

    return ul_tag, div_tag


def wrap_each_content(
    contents: list[TagChild], wrapper: WrapperCallable = card_body
) -> list[CardItem]:
    ret: list[CardItem] = []
    for content_item in contents:
        if isinstance(content_item, CardItem):
            ret.append(content_item)
        else:
            ret.append(wrapper(content_item))
    return ret


def navset_title(
    title: TagChild | None,
) -> list[TagChild | TagAttrs]:
    """Note: Return value should be spread into parent container."""

    if title is None:
        return [None]

    title_attrs: TagAttrs = {"class": "bslib-navs-card-title"}
    return [title_attrs, tags.span(title)]


def nav_random_int() -> str:
    return private_random_int(1000, 1000000)



from shiny import reactive
from shiny.express import input, ui

with ui.sidebar():
    ui.input_radio_buttons("controller", "Controller", ["1", "2", "3"], selected="1")

with ui.navset_hidden(id="hidden_tabs"):
    with ui.nav_panel(None, value="panel1"):
        "Panel 1 content"
    with ui.nav_panel(None, value="panel2"):
        "Panel 2 content"
    with ui.nav_panel(None, value="panel3"):
        "Panel 3 content"


@reactive.effect
@reactive.event(input.controller)
def _():
    ui.update_navs("hidden_tabs", selected="panel" + str(input.controller()))



from ._fill import (
    as_fill_item,
    as_fillable_container,
    remove_all_fill,
    # is_fill_item,
    # is_fillable_container,
)

__all__ = (
    "as_fillable_container",
    "as_fill_item",
    "remove_all_fill",
    # "is_fillable_container",
    # "is_fill_item",
)



from __future__ import annotations

from typing import Any, Dict, cast

from htmltools import TagFunction

from ...session._utils import RenderedDeps
from ...types import MISSING_TYPE, ImgData
from ._renderer import Jsonifiable

JsonifiableDict = Dict[str, Jsonifiable]


def rendered_deps_to_jsonifiable(rendered_deps: RenderedDeps) -> JsonifiableDict:
    return cast(JsonifiableDict, dict(rendered_deps))


def imgdata_to_jsonifiable(imgdata: ImgData) -> JsonifiableDict:
    return cast(JsonifiableDict, dict(imgdata))


def set_kwargs_value(
    kwargs: dict[str, Any],
    key: str,
    ui_val: TagFunction | str | float | int | MISSING_TYPE,
    self_val: TagFunction | str | float | int | None | MISSING_TYPE,
):
    """
    Set kwarg value with fallback value.

    * If `ui_val` is not `MISSING`, set `kwargs[key] = ui_val`.
    * If `self_val` is not `MISSING` and is not `None`, set `kwargs[key] = self_val`.
    * Otherwise, do nothing.
    """
    if not isinstance(ui_val, MISSING_TYPE):
        kwargs[key] = ui_val
        return
    if not (isinstance(self_val, MISSING_TYPE) or self_val is None):
        kwargs[key] = self_val
        return
    # Do nothing as we don't want to override the default value (that could change in the future)
    return



import functools
from typing import Callable, TypeVar

from .._docstring import add_example
from .._typing_extensions import Concatenate, ParamSpec
from ..module import Id
from ..session._session import Inputs, Outputs, Session
from ..session._utils import require_active_session, session_context
from .expressify_decorator import expressify

T = TypeVar("T")
P = ParamSpec("P")
R = TypeVar("R")

__all__ = ("module",)


@add_example(ex_dir="../api-examples/express_module")
def module(
    fn: Callable[Concatenate[Inputs, Outputs, Session, P], R]
) -> Callable[Concatenate[Id, P], R]:
    """
    Create a Shiny module using Shiny Express syntax

    This function is used to create a Shiny module, where the code inside the function
    uses Shiny Express syntax. This is in contrast to the pair of functions
    :func:`~shiny.module.ui()` and :func:`~shiny.module.server()`, which are used to
    create Shiny modules with Core syntax.

    Parameters
    ----------
    fn
        The function that defines the module. The first three parameters of this
        function must be `input`, `output`, and `session`. Any additional parameters can
        used to pass information to the module.

    See Also
    --------
    * ~shiny.module.ui
    * ~shiny.module.server
    * ~shiny.express.expressify
    """
    fn = expressify(fn)

    @functools.wraps(fn)
    def wrapper(id: Id, *args: P.args, **kwargs: P.kwargs) -> R:
        parent_session = require_active_session(None)
        module_session = parent_session.make_scope(id)

        with session_context(module_session):
            return fn(
                module_session.input,
                module_session.output,
                module_session,
                *args,
                **kwargs,
            )

    return wrapper



from shiny.express import input, render, ui

ui.page_opts(title="Checkbox Kitchen Sink", fillable=True)

with ui.layout_columns():
    with ui.card():
        ui.card_header("Default checkbox with label")
        ui.input_checkbox("default", "Basic Checkbox")

        @render.code
        def default_txt():
            return str(input.default())

    with ui.card():
        ui.card_header("Checkbox With Value")
        ui.input_checkbox("value", "Checkbox with Value", value=True)

        @render.code
        def value_txt():
            return str(input.value())

    with ui.card():
        ui.card_header("Checkbox With Width")
        ui.input_checkbox("width", "Checkbox with Width", width="10px")

        @render.code
        def width_txt():
            return str(input.width())



dark_color, light_color = [
    ("#1A2A6C", "#AED9E0"),  # Fix to `blue` as it is a e2e app, and not a demo
    ("#800020", "#F6DFD7"),
    ("#4B0082", "#E6E6FA"),
    ("#006D5B", "#A2D5C6"),
][0]

adjectives = [
    "charming",
    "cuddly",
    "elegant",
    "fierce",
    "graceful",
    "majestic",
    "playful",
    "quirky",
    "silly",
    "witty",
]

animals = [
    "elephant",
    "giraffe",
    "jaguar",
    "koala",
    "lemur",
    "otter",
    "panda",
    "panther",
    "penguin",
    "zebra",
]



from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.playwright.expect._internal import _expect_nav_to_have_header_footer
from shiny.run import ShinyAppProc


def test_navset_hidden_kitchensink(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    navset_hidden_1 = controller.NavsetHidden(page, "hidden_tabs1")
    radio1 = controller.InputRadioButtons(page, "controller1")
    navset_hidden_1.expect_value("panel2")
    navset_hidden_1._expect_content_text("Panel 2 content")
    # assert the DOM structure for hidden_navset with header and footer is preserved
    _expect_nav_to_have_header_footer(
        navset_hidden_1.get_loc_active_content().locator("..").locator(".."),
        "navset_hidden_header1",
        "navset_hidden_footer1",
    )
    # assert header and footer contents
    expect(page.locator("#navset_hidden_header1")).to_contain_text(
        "Navset_hidden_header"
    )
    expect(page.locator("#navset_hidden_footer1")).to_contain_text(
        "Navset_hidden_footer"
    )
    radio1.set("1")
    navset_hidden_1.expect_value("panel1")
    navset_hidden_1._expect_content_text("Panel 1 content")

    navset_hidden_2 = controller.NavsetHidden(page, "hidden_tabs2")
    navset_hidden_2.expect_value("panel4")



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def expect_row_selection(page: Page, prefix_main: str, prefix_secondary: str):
    # TODO: Karan, add Dataframe helpers methods for selected_rows

    main_txt_rows = controller.OutputTextVerbatim(page, f"{prefix_main}selected_rows")
    other_txt_rows = controller.OutputTextVerbatim(
        page, f"{prefix_secondary}selected_rows"
    )

    main_selected_row_count = controller.OutputTextVerbatim(
        page, f"{prefix_main}selected_row_count"
    )
    other_selected_row_count = controller.OutputTextVerbatim(
        page, f"{prefix_secondary}selected_row_count"
    )

    main_grid_row_count = controller.OutputTextVerbatim(
        page, f"{prefix_main}grid_row_count"
    )
    other_grid_row_count = controller.OutputTextVerbatim(
        page, f"{prefix_secondary}grid_row_count"
    )

    main_select_btn = controller.InputActionButton(page, f"{prefix_main}select")
    main_clear_btn = controller.InputActionButton(page, f"{prefix_main}clear")

    main_txt_rows.expect_value("()")
    other_txt_rows.expect_value("()")
    main_selected_row_count.expect_value("0")
    other_selected_row_count.expect_value("0")
    main_grid_row_count.expect_value("344")
    other_grid_row_count.expect_value("344")

    main_select_btn.click()
    main_txt_rows.expect_value("(1, 3, 5)")
    other_txt_rows.expect_value("()")
    main_selected_row_count.expect_value("3")
    other_selected_row_count.expect_value("0")
    main_grid_row_count.expect_value("344")
    other_grid_row_count.expect_value("344")

    main_clear_btn.click()

    main_txt_rows.expect_value("()")
    other_txt_rows.expect_value("()")
    main_selected_row_count.expect_value("0")
    other_selected_row_count.expect_value("0")
    main_grid_row_count.expect_value("344")
    other_grid_row_count.expect_value("344")


def test_row_selection(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)
    expect_row_selection(page, "", "pandas-")
    expect_row_selection(page, "pandas-", "")

    expect_row_selection(page, "", "polars-")
    expect_row_selection(page, "polars-", "")



from __future__ import annotations

import pkgutil

# pyright: reportUnknownMemberType = false
# pyright: reportMissingTypeStubs = false
# pyright: reportArgumentType = false
# pyright: reportUnknownMemberType = false
#
# TODO-barret-render.data_frame; Make an example that uses a dataframe that then updates a higher level reactive, that causes the df to update... which causes the table to render completely
# TODO-barret-render.data_frame; When "updating" data, try to maintain the scroll, filter info when a new `df` is supplied;
#
# TODO-karan-test; Click outside the table. Tab to the column name, hit enter. Verify the table becomes sorted. Tab to an HTML column name, hit enter. Verify the sort does not update.
#
# TODO-future; Can we maintain pre-processed value and use it within editing?
# A: Doesn't seem possible for now
import great_tables as gt
import palmerpenguins  # pyright: ignore[reportMissingTypeStubs]
import polars as pl

from shiny import App, Inputs, Outputs, Session, module, reactive, render, req, ui
from shiny.render import CellPatch
from shiny.render._data_frame_utils._styles import StyleInfo
from shiny.types import Jsonifiable

pd_penguins = palmerpenguins.load_penguins_raw()
pl_penguins = pl.read_csv(
    pkgutil.get_data("palmerpenguins", "data/penguins-raw.csv"),
    null_values="NA",
)

pd_asdf = render.DataTable(pd_penguins)
pl_asdf = render.DataTable(pl_penguins)

# Load the dataset
df = pd_penguins

df = df.head(15)


# Add some HTML content to the dataframe!
df = df.astype({"Sample Number": "object"})
# df.loc[:, "Sample Number"] = df.loc[:, "Sample Number"].astype("object")
df.loc[:, "Sample Number"] = df.loc[
    :, "Sample Number"
].apply(  # pyright: ignore[reportCallIssue]
    lambda x: ui.HTML(  # pyright: ignore[reportUnknownLambdaType]
        str(
            ui.tags.strong(
                ui.tags.em(str(x))  # pyright: ignore[reportUnknownArgumentType]
            )
        )
    )
)


# Use a non-standard index, just in case
df.loc[:, "test_index"] = [
    f"Row {i}" for i in range(0, df.shape[0])
]  # pyright: ignore[reportUnknownArgumentType]
df.set_index("test_index", drop=True, inplace=True)


# print(p)
df = pl_penguins.head(15).with_columns(
    (pl.Series([ui.tags.strong(ui.tags.em(str(x))) for x in range(0, 15)])).alias(
        "Sample Number"
    )
)


print(df)


def gt_style_str_to_obj(style_str: str) -> dict[str, Jsonifiable]:
    # Could use BeautifulSoup here, but this is a simple example
    style_obj: dict[str, Jsonifiable] = {}
    for style_part in style_str.split(";"):
        style_part = style_part.strip()
        if not style_part:
            continue
        key, value = style_part.split(":")
        style_obj[key] = value
    return style_obj


def gt_styles(df_gt: gt.GT) -> list[StyleInfo]:
    styles = df_gt._styles
    ret: list[StyleInfo] = []
    for style in styles:
        location = style.locname
        location = (
            "body"
            if location == "data"  # pyright: ignore[reportUnnecessaryComparison]
            else location
        )
        assert location == "body", f"`style.locname` is {location}, expected 'body'"
        rows = style.rownum
        assert rows is not None
        cols = style.colnum
        if cols is None:
            cols = style.colname
        assert cols is not None

        style_obj: dict[str, Jsonifiable] = {}
        for style_str in style.styles:
            style_obj.update(
                gt_style_str_to_obj(style_str._to_html_style()),
            )
        ret.append(
            {
                "location": location,
                "rows": rows,
                "cols": cols,
                "style": style_obj,
            }
        )
    return ret


# import numpy as np
# import pandas as pd

# np.random.seed(0)
# df2 = pd.DataFrame(np.random.randn(10, 4), columns=["A", "B", "C", "D"])
# df2.style

# df = df2.style
# print(df, dir(df))


MOD_ID = "testing"


@module.ui
def mod_ui():
    return ui.TagList(
        ui.card(
            ui.output_data_frame("selected_summary_data"),
            ui.output_data_frame("summary_data"),
            height="400px",
        ),
    )


app_ui = ui.page_fillable(
    {"class": "p-3"},
    # ui.markdown(
    #     "**Instructions**: Select one or more countries in the table below to see more information."
    # ),
    mod_ui(MOD_ID),
)


@module.server
def mod_server(input: Inputs, output: Outputs, session: Session):

    @render.data_frame
    def selected_summary_data():
        return summary_data.data_view(selected=True)

    @reactive.effect
    async def _():
        req(summary_data.data_view_rows())
        await summary_data.update_cell_selection(
            {
                "type": "row",
                "rows": [0, 2],
            }
        )

    @render.data_frame
    def summary_data():

        df_gt = (
            gt.GT(df)
            .tab_style(
                [
                    gt.style.fill(color="purple"),
                    gt.style.borders(color="green", style="dashed"),
                ],
                gt.loc.body("Species", [1, 2]),
            )
            .tab_style(
                gt.style.fill(color="yellow"),
                gt.loc.body("Region", [2]),
            )
            .tab_style(
                gt.style.fill(color="red"),
                gt.loc.body("Island"),
            )
        )
        df_styles = gt_styles(df_gt)
        counter = 0

        import pandas as pd

        def df_styles_fn(data: pd.DataFrame) -> list[StyleInfo]:
            nonlocal counter
            counter = counter + 2
            if counter > 5:
                counter = 0

            ret: list[StyleInfo] = []
            while len(ret) < counter and len(ret) < len(df_styles):
                ret.append(df_styles[len(ret)])

            return ret

        # NOTE - Styles in GT are greedy!
        # Q: Cell editing should not be allowed as styles would be applied given the original data
        #   * Once GT becomes not greedy, we can unlock this.
        # Styles can be subsetted by looking for the row value
        # return df
        return render.DataTable(
            df,
            selection_mode=("rows"),
            editable=True,
            # filters=True,
            styles=df_styles_fn,
        )
        # Ideal
        return render.DataGrid(
            df_gt,
            # selection_mode=("rows"),
            # editable=True,
            filters=True,
        )
        # return render.DataTable(df, selection_mode="none", editable=True)
        return render.DataGrid(df, selection_mode="rows", editable=True)
        # return render.DataTable(df, selection_mode="rows", editable=True)
        # return render.DataGrid(df, selection_mode="rows", editable=False)
        # return render.DataTable(df, selection_mode="rows", editable=False)

    # @reactive.effect
    # def _():
    #     print("Filters:", summary_data.filter())
    # @reactive.effect
    # def _():
    #     print("Sorting:", summary_data.sort())
    # @reactive.effect
    # def _():
    #     print("indices:", summary_data.data_view_rows())
    # @reactive.effect
    # def _():
    #     print("Data View:\n", summary_data.data_view(selected=False))
    # @reactive.effect
    # def _():
    #     print("Data View (selected):\n", summary_data.data_view(selected=True))
    # @reactive.effect
    # def _():
    #     print("Cell Selection:", summary_data.cell_selection())
    # @reactive.effect
    # def _():
    #     print(summary_data.data())
    #     print(summary_data.data_patched())
    #     print(summary_data.cell_patches())
    # from shiny import reactive
    # @reactive.effect
    # def _():
    #     print(summary_data._type_hints())
    from shinywidgets import render_widget

    # from shiny import reactive

    @render_widget
    def country_detail_pop():
        import plotly.express as px

        return px.line(
            px.data.gapminder(),
            x="year",
            y="lifeExp",
            color="country",
            title="Population Over Time",
        )

    @summary_data.set_patch_fn
    def upgrade_patch(
        *,
        patch: CellPatch,
    ):
        # from shinywidgets import output_widget

        # if len(summary_data.cell_patches()) == 0:
        #     return ui.card(output_widget("country_detail_pop"), height="400px")

        # import time

        # time.sleep(2)
        if len(summary_data.cell_patches()) > 3:
            import random

            from shiny.types import SafeException

            if random.sample([0, 1], 1)[0]:
                raise SafeException("Safe error message!")
            else:
                raise RuntimeError("Barret testing!")

        val = ui.HTML(f"demo_{patch['value']}")
        # return val

        # print(str(len(summary_data.cell_patches())))
        import random

        from htmltools import HTMLDependency, TagList

        return TagList(
            HTMLDependency(
                "barret" + "".join(random.choices("0123456789", k=5)),
                version="1",
                head=f"""
                <script>console.log("patching: {val}")</script>
                """,
            ),
            val,
        )

        return ui.HTML(
            f"""
            <script>console.log("patching: {patch['value']}")</script>
            demo_{patch["value"]}
            """
        )

    # @reactive.effect
    # def _():
    #     print(session.id)
    #     print(summary_data.data_patched())
    #     print("patches: ", summary_data.cell_patches())

    if False:

        @summary_data.set_patches_fn
        async def upgrade_patches(
            # data: pd.DataFrame,
            *,
            patches: list[CellPatch],
            # row_index: int,
            # column_id: str,
            # value: str,
        ):

            new_patches: list[CellPatch] = []
            for patch in patches:
                # prev_value = summary_data.data()
                new_patches.append(
                    {
                        "row_index": patch["row_index"],
                        "column_index": patch["column_index"],
                        "value": await summary_data._patch_fn(patch=patch),
                    }
                )
                extra_patch = patch.copy()
                extra_patch["column_index"] = abs(patch["column_index"] - 1)
                new_patches.append(
                    {
                        "row_index": extra_patch["row_index"],
                        "column_index": extra_patch["column_index"],
                        "value": await summary_data._patch_fn(patch=extra_patch),
                    }
                )

            return new_patches


def server(input: Inputs, output: Outputs, session: Session):
    mod_server(MOD_ID)


app = App(app_ui, server, debug=False)
app.sanitize_errors = True



from conftest import create_doc_example_core_fixture
from playwright.sync_api import Page, expect

from shiny.playwright import controller
from shiny.run import ShinyAppProc

app = create_doc_example_core_fixture("input_checkbox_group")


def test_input_checkbox_group_kitchen(page: Page, app: ShinyAppProc) -> None:
    page.goto(app.url)

    colors = controller.InputCheckboxGroup(page, "colors")

    expect(colors.loc_label).to_have_text("Choose color(s):")
    colors.expect_label("Choose color(s):")

    colors.expect_choices(["red", "green", "blue"])
    colors.expect_choice_labels(["Red", "Green", "Blue"])
    colors.expect_selected([])
    colors.expect_inline(False)

    colors.expect_width(None)

    colors.set(["red", "blue"])

    colors.expect_choices(["red", "green", "blue"])
    colors.expect_choice_labels(["Red", "Green", "Blue"])
    colors.expect_selected(["red", "blue"])



# TODO-barret; Test this app!

from __future__ import annotations

import palmerpenguins  # pyright: ignore[reportMissingTypeStubs]
import polars as pl

from shiny import reactive
from shiny.express import input, render, ui

pd_data = palmerpenguins.load_penguins_raw().iloc[0:2, 0:2]
pl_data = pl.DataFrame(pd_data)

with ui.layout_column_wrap(width=1 / 2):

    with ui.card():

        ui.h3("Pandas")

        ui.input_action_button("update_pd_btn", "Update cell")

        @render.data_frame
        def pd_df():
            return pd_data

    @reactive.effect
    @reactive.event(input.update_pd_btn)
    async def _():

        await pd_df.update_cell_value(
            "pandas - " + str(input.update_pd_btn()),
            row=0,
            col=0,
        )

    with ui.card():

        ui.h3("Polars")
        ui.input_action_button("update_pl_btn", "Update cell")

        @render.data_frame
        def pl_df():
            return pl_data

    @reactive.effect
    @reactive.event(input.update_pl_btn)
    async def _():

        await pl_df.update_cell_value(
            "polars - " + str(input.update_pl_btn()),
            row=0,
            col=0,
        )



"""This script is NOT designed to be run as part of a pytest suite; it demands a clean
Python process because it needs to be run before the shiny module is loaded.

The purpose of this script is to ensure that no shiny module initialization code uses
the event loop. The running event loop during shiny module startup is different than the
running event loop during the app's operation (I assume uvicorn creates the latter). The
creation of, say, an asyncio.Lock during the course of shiny module startup will result
in race conditions if the lock is used during the app's operation."""

if __name__ == "__main__":
    import asyncio
    import importlib
    import sys
    from typing import Optional

    if "shiny" in sys.modules:
        raise RuntimeError(
            "Bad test: shiny was already loaded, it's important that SpyEventLoopPolicy"
            " is installed before shiny loads"
        )

    class SpyEventLoopPolicy(asyncio.DefaultEventLoopPolicy):
        def get_event_loop(self):
            raise RuntimeError("get_event_loop called")

        def set_event_loop(self, loop: Optional[asyncio.AbstractEventLoop]):
            raise RuntimeError("set_event_loop called")

        def new_event_loop(self):
            raise RuntimeError("new_event_loop called")

    asyncio.set_event_loop_policy(SpyEventLoopPolicy())

    # Doing this instead of "import shiny" so no linter is tempted to remove it
    importlib.import_module("shiny")
    print(
        "Success; shiny module loading did not attempt to access an asyncio event "
        "loop\n",
        file=sys.stderr,
    )



import shiny.experimental as x
from shiny import App, ui

app_ui = ui.page_fluid(
    ui.tags.style(
        """
        .card-body {
            border: 1px dashed red;
        }
        """
    ),
    x.ui.card(
        ui.card_header(
            "These two images are in their own individual card_body() container"
        ),
        x.ui.card_image(
            file=None,
            src="https://posit.co/wp-content/uploads/2022/10/Posit-logo-h-full-color-RGB-TM.svg",
        ),
        x.ui.card_image(
            file=None,
            src="https://posit.co/wp-content/uploads/2022/10/Posit-logo-h-full-color-RGB-TM.svg",
        ),
        ui.card_header("These two images are in the same card_body() container"),
        x.ui.card_image(
            file=None,
            src="https://posit.co/wp-content/uploads/2022/10/Posit-logo-h-full-color-RGB-TM.svg",
            container=ui.tags.span,
        ),
        x.ui.card_image(
            file=None,
            src="https://posit.co/wp-content/uploads/2022/10/Posit-logo-h-full-color-RGB-TM.svg",
            container=ui.tags.span,
        ),
    ),
)


app = App(app_ui, server=None)



from shiny import App, Inputs, Outputs, Session, module, ui


def my_ui(message: str) -> ui.TagList:
    return ui.TagList(
        ui.input_checkbox("show", "Show secret message", False),
        ui.panel_conditional(
            "input.show",
            message,
            id=module.resolve_id("cond_message"),
        ),
    )


mod_ui = module.ui(my_ui)


@module.server
def mod_server(input: Inputs, output: Outputs, session: Session): ...


app_ui = ui.page_fluid(
    ui.h3("Non-module version"),
    my_ui("Lorem ipsum dolor sit amet"),
    ui.hr(),
    ui.h3("Module version"),
    mod_ui("mod", "consectetur adipiscing elit"),
)


def server(input: Inputs, output: Outputs, session: Session):
    mod_server("mod")


app = App(app_ui, server)



import random

from shiny import App, Inputs, Outputs, Session, reactive, render, ui

app_ui = ui.page_fluid(ui.output_text("value"))


def server(input: Inputs, output: Outputs, session: Session):
    @render.text
    def value():
        reactive.invalidate_later(0.5)
        return "Random int: " + str(random.randint(0, 10000))


app = App(app_ui, server)



import matplotlib.pyplot as plt
import numpy as np

from shiny import App, Inputs, Outputs, Session, render, ui

app_ui = ui.page_fluid(
    ui.input_slider("obs", "Number of bins:", min=10, max=100, value=30),
    ui.output_plot("distPlot"),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.plot
    def distPlot():
        np.random.seed(19680801)
        x = 100 + 15 * np.random.randn(437)

        fig, ax = plt.subplots()
        ax.hist(x, input.obs(), density=True)
        return fig


app = App(app_ui, server)



# ------------------------------------------------------------------------------------
# When putting a Chat into production, there are at least a couple additional
# considerations to keep in mind:
#  - Token Limits: LLMs have (varying) limits on how many tokens can be included in
#    a single request and response. To accurately respect these limits, you'll want
#    to find the revelant limits and tokenizer for the model you're using, and inform
#    Chat about them.
#  - Reproducibility: Consider pinning a snapshot of the LLM model to ensure that the
#    same model is used each time the app is run.
#
# See the MODEL_INFO dictionary below for an example of how to set these values for
# Anthropic's Claude model.
# https://docs.anthropic.com/en/docs/about-claude/models#model-comparison-table
# ------------------------------------------------------------------------------------
import os

from anthropic import AsyncAnthropic
from app_utils import load_dotenv

from shiny.express import ui

load_dotenv()
llm = AsyncAnthropic(api_key=os.environ.get("ANTHROPIC_API_KEY"))


MODEL_INFO = {
    "name": "claude-3-5-sonnet-20241022",
    # DISCLAIMER: Anthropic has not yet released a public tokenizer for Claude models,
    # so this uses the generic default provided by Chat() (for now). That is probably
    # ok though since the default tokenizer likely overestimates the token count.
    "tokenizer": None,
    "token_limits": (200000, 8192),
}


ui.page_opts(
    title="Hello Anthropic Chat",
    fillable=True,
    fillable_mobile=True,
)

chat = ui.Chat(
    id="chat",
    messages=[
        {"content": "Hello! How can I help you today?", "role": "assistant"},
    ],
    tokenizer=MODEL_INFO["tokenizer"],
)

chat.ui()


@chat.on_user_submit
async def _():
    messages = chat.messages(
        format="anthropic",
        token_limits=MODEL_INFO["token_limits"],
    )
    response = await llm.messages.create(
        model=MODEL_INFO["name"],
        messages=messages,
        stream=True,
        max_tokens=MODEL_INFO["token_limits"][1],
    )
    await chat.append_message_stream(response)



from __future__ import annotations

from typing import Optional

import pytest

from shiny import reactive, render
from shiny.render.renderer import Renderer, ValueFn
from shiny.session import Session, get_current_session


class RendererWithSession(Renderer[str]):
    _session: Session | None

    def __init__(self, _fn: Optional[ValueFn[str]] = None):
        super().__init__(_fn)
        self._session = get_current_session()


@pytest.mark.asyncio
async def test_renderer_works():
    # No args works
    class test_renderer(RendererWithSession):
        async def transform(self, value: str) -> str:
            return value + " " + value

    @test_renderer()
    def txt_paren() -> str:
        return "Hello World!"

    val = await txt_paren.render()
    assert val == "Hello World! Hello World!"

    @test_renderer
    def txt_no_paren() -> str:
        return "Hello World!"

    val = await txt_no_paren.render()
    assert val == "Hello World! Hello World!"


@pytest.mark.asyncio
async def test_renderer_works_with_args():
    # No args works
    class test_renderer_with_args(RendererWithSession):
        def __init__(self, _fn: Optional[ValueFn[str]] = None, *, times: int = 2):
            super().__init__(_fn)
            self.times: int = times

        async def transform(self, value: str) -> str:
            values = [value for _ in range(self.times)]
            return " ".join(values)

    @test_renderer_with_args
    def txt2() -> str:
        return "42"

    @test_renderer_with_args(times=4)
    def txt4() -> str:
        return "42"

    val = await txt2.render()
    assert val == "42 42"
    val = await txt4.render()
    assert val == "42 42 42 42"


def test_effect():
    with pytest.raises(TypeError):

        @reactive.effect  # pyright: ignore[reportArgumentType]
        @render.text
        def my_output():
            return "42"



from shiny import reactive
from shiny.express import input, ui

ui.input_action_button("btn_show", "Show popover", class_="mt-3 me-3")
ui.input_action_button("btn_close", "Close popover", class_="mt-3 me-3")

ui.br()
ui.br()

with ui.popover(id="popover_id"):
    ui.input_action_button("btn_w_popover", "A button w/ a popover", class_="mt-3")
    "A message"


@reactive.effect
@reactive.event(input.btn_show)
def _():
    ui.update_popover("popover_id", show=True)


@reactive.effect
@reactive.event(input.btn_close)
def _():
    ui.update_popover("popover_id", show=False)


@reactive.effect
@reactive.event(input.btn_w_popover)
def _():
    ui.notification_show("Button clicked!", duration=3, type="message")



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_card(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    card = controller.Card(page, "card1")
    card.expect_max_height(None)
    card.expect_min_height(None)
    card.expect_height(None)
    card.expect_header("This is the header")
    card.expect_footer("This is the footer")
    card.expect_body(
        [
            "\nThis is the title\nThis is the body.\n",
            "\n\n",
            "\nThis is still the body.\n",
        ]
    )
    card.expect_full_screen(False)
    card.set_full_screen(True)
    card.expect_full_screen(True)
    card.set_full_screen(False)
    card.expect_full_screen(False)



import seaborn as sns

# Import data from shared.py
from shared import df

from shiny import App, render, ui

app_ui = ui.page_sidebar(
    ui.sidebar(
        ui.input_select(
            "var", "Select variable", choices=["bill_length_mm", "body_mass_g"]
        ),
        ui.input_switch("species", "Group by species", value=True),
        ui.input_switch("show_rug", "Show Rug", value=True),
    ),
    ui.output_plot("hist"),
    title="Hello sidebar!",
)


def server(input, output, session):
    @render.plot
    def hist():
        hue = "species" if input.species() else None
        sns.kdeplot(df, x=input.var(), hue=hue)
        if input.show_rug():
            sns.rugplot(df, x=input.var(), hue=hue, color="black", alpha=0.25)


app = App(app_ui, server)



from __future__ import annotations

import copy
import os
import pathlib
import shutil
import tempfile
from typing import BinaryIO, List, Optional, cast

from . import _utils
from .types import FileInfo

# File uploads happen through a series of requests. This requires a browser
# which supports the HTML5 File API.
#
# 1. Client tells server that one or more files are about to be uploaded, with
#    an "uploadInit" message; the server responds with a "jobId" and "uploadUrl"
#    that the client should use to upload the files. From the server's
#    perspective, the messages look like this:
#    RECV {"method":"uploadInit","args":[[{"name":"mtcars.csv","size":1303,"type":"text/csv"}]],"tag":2}
#    SEND {"response":{"tag":2,"value":{"jobId":"1651ddebfb643a26e6f18aa1","uploadUrl":"session/3cdbe3c4d1318225fee8f2e3417a1c99/upload/1651ddebfb643a26e6f18aa1?w="}}}
#
# 2. For each file (sequentially):
#    b. Client makes a POST request with the file data.
#    c. Server sends a 200 response to the client.
#
# 3. Repeat 2 until all files have been uploaded.
#
# 4. Client tells server that all files have been uploaded, along with the
#    input ID that this data should be associated with. The server responds
#    with the tag ID and a null message. The messages look like this:
#    RECV {"method":"uploadEnd","args":["1651ddebfb643a26e6f18aa1","file1"],"tag":3}
#    SEND {"response":{"tag":3,"value":null}}


class FileUploadOperation:
    def __init__(
        self, parent: FileUploadManager, id: str, dir: str, file_infos: List[FileInfo]
    ) -> None:
        self._parent: FileUploadManager = parent
        self._id: str = id
        self._dir: str = dir
        # Copy file_infos and add a "datapath" entry for each file.
        self._file_infos: list[FileInfo] = [
            cast(FileInfo, {**fi, "datapath": ""}) for fi in copy.deepcopy(file_infos)
        ]
        self._n_uploaded: int = 0
        self._current_file_obj: Optional[BinaryIO] = None

    # Start uploading one of the files.
    def file_begin(self) -> None:
        file_info: FileInfo = self._file_infos[self._n_uploaded]
        file_ext = pathlib.Path(file_info["name"]).suffix
        file_info["datapath"] = os.path.join(
            self._dir, str(self._n_uploaded) + file_ext
        )
        self._current_file_obj = open(file_info["datapath"], "ab")

    # Finish uploading one of the files.
    def file_end(self) -> None:
        if self._current_file_obj is not None:
            self._current_file_obj.close()
        self._current_file_obj = None
        self._n_uploaded += 1

    # Write a chunk of data for the currently-open file.
    def write_chunk(self, chunk: bytes) -> None:
        if self._current_file_obj is None:
            raise RuntimeError(f"FileUploadOperation for {self._id} is not open.")
        self._current_file_obj.write(chunk)

    # End the entire operation, which can consist of multiple files.
    def finish(self) -> List[FileInfo]:
        if self._n_uploaded != len(self._file_infos):
            raise RuntimeError(
                f"Not all files for FileUploadOperation {self._id} were uploaded."
            )
        self._parent.on_job_finished(self._id)
        return self._file_infos

    # Context handlers for `with`
    def __enter__(self) -> None:
        self.file_begin()

    def __exit__(self, type, value, trace) -> None:  # type: ignore
        self.file_end()


class FileUploadManager:
    def __init__(self) -> None:
        # TODO: Remove basedir when app exits.
        self._basedir: str = tempfile.mkdtemp(prefix="fileupload-")
        self._operations: dict[str, FileUploadOperation] = {}

    def create_upload_operation(self, file_infos: List[FileInfo]) -> str:
        job_id = _utils.rand_hex(12)
        dir = tempfile.mkdtemp(dir=self._basedir)
        self._operations[job_id] = FileUploadOperation(self, job_id, dir, file_infos)
        return job_id

    def get_upload_operation(self, id: str) -> Optional[FileUploadOperation]:
        if id in self._operations:
            return self._operations[id]
        else:
            return None

    def on_job_finished(self, job_id: str) -> None:
        del self._operations[job_id]

    # Remove the directories containing file uploads; this is to be called when
    # a session ends.
    def rm_upload_dir(self) -> None:
        shutil.rmtree(self._basedir)



import starlette.responses

from shiny import App, Inputs, Outputs, Session, render, ui

app_ui = ui.page_fluid(
    ui.markdown(
        """
        ## Sticky load balancing test

        The purpose of this app is to determine if HTTP requests made by the client are
        correctly routed back to the same Python process where the session resides. It
        is only useful for testing deployments that load balance traffic across more
        than one Python process.

        If this test fails, it means that sticky load balancing is not working, and
        certain Shiny functionality (like file upload/download or server-side selectize)
        are likely to randomly fail.
        """
    ),
    ui.tags.div(
        {"class": "card"},
        ui.tags.div(
            {"class": "card-body font-monospace"},
            ui.tags.div("Attempts: ", ui.tags.span("0", id="count")),
            ui.tags.div("Status: ", ui.tags.span(id="status")),
            ui.output_ui("out"),
        ),
    ),
)


def server(input: Inputs, output: Outputs, session: Session):
    @render.ui
    def out():
        # Register a dynamic route for the client to try to connect to.
        # It does nothing, just the 200 status code is all that the client
        # will care about.
        url = session.dynamic_route(
            "test",
            lambda req: starlette.responses.PlainTextResponse(
                "OK", headers={"Cache-Control": "no-cache"}
            ),
        )

        # Send JS code to the client to repeatedly hit the dynamic route.
        # It will succeed if and only if we reach the correct Python
        # process.
        return ui.tags.script(
            f"""
            const url = "{url}";
            const count_el = document.getElementById("count");
            const status_el = document.getElementById("status");
            let count = 0;
            async function check_url() {{
                count_el.innerHTML = ++count;
                try {{
                    const resp = await fetch(url);
                    if (!resp.ok) {{
                        status_el.innerHTML = "Failure!";
                        return;
                    }} else {{
                        status_el.innerHTML = "In progress";
                    }}
                }} catch(e) {{
                    status_el.innerHTML = "Failure!";
                    return;
                }}

                if (count === 100) {{
                    status_el.innerHTML = "Test complete";
                    return;
                }}

                setTimeout(check_url, 10);
            }}
            check_url();
            """
        )


app = App(app_ui, server)



from ._input_buttons import (
    InputActionButton,
    InputActionLink,
    InputDarkMode,
    InputFile,
    InputTaskButton,
)

from ._input_fields import (
    InputDate,
    InputDateRange,
    InputNumeric,
    InputPassword,
    InputText,
    InputTextArea,
)

from ._input_controls import (
    InputCheckbox,
    InputCheckboxGroup,
    InputRadioButtons,
    InputSelect,
    InputSelectize,
    InputSlider,
    InputSliderRange,
    InputSwitch,
)

from ._overlay import (
    Popover,
    Tooltip,
)

from ._layout import (
    Sidebar,
)

from ._accordion import (
    Accordion,
    AccordionPanel,
)

from ._card import Card, ValueBox

from ._file import (
    DownloadButton,
    DownloadLink,
)

from ._chat import (
    Chat,
)
from ._navs import (
    NavPanel,
    NavsetBar,
    NavsetCardPill,
    NavsetCardTab,
    NavsetCardUnderline,
    NavsetHidden,
    NavsetPill,
    NavsetPillList,
    NavsetTab,
    NavsetUnderline,
    PageNavbar,
)
from ._output import (
    OutputCode,
    OutputDataFrame,
    OutputImage,
    OutputPlot,
    OutputTable,
    OutputText,
    OutputTextVerbatim,
    OutputUi,
)

__all__ = [
    "InputActionButton",
    "InputActionLink",
    "InputCheckbox",
    "InputCheckboxGroup",
    "InputDarkMode",
    "InputDate",
    "InputDateRange",
    "InputFile",
    "InputNumeric",
    "InputPassword",
    "InputRadioButtons",
    "InputSelect",
    "InputSelectize",
    "InputSlider",
    "InputSliderRange",
    "InputSwitch",
    "InputTaskButton",
    "InputText",
    "InputTextArea",
    "OutputCode",
    "OutputDataFrame",
    "OutputImage",
    "OutputPlot",
    "OutputTable",
    "OutputText",
    "OutputTextVerbatim",
    "OutputUi",
    "ValueBox",
    "Card",
    "Chat",
    "Accordion",
    "AccordionPanel",
    "Sidebar",
    "Popover",
    "Tooltip",
    "NavPanel",
    "NavsetBar",
    "NavsetCardPill",
    "NavsetCardTab",
    "NavsetCardUnderline",
    "NavsetHidden",
    "NavsetPill",
    "NavsetPillList",
    "NavsetTab",
    "NavsetUnderline",
    "DownloadButton",
    "DownloadLink",
    "PageNavbar",
]



#!/usr/bin/env python3

import importlib
import sys
from types import ModuleType
from typing import Union, cast

if sys.version_info < (3, 9):
    raise RuntimeError("This script requires Python 3.9 or later")


def gen_init_py(package_name: str) -> None:
    """
    Generate imports for __init__.py for a package or subpackage.
    """
    pkg = importlib.import_module(package_name)
    print("")
    print("=" * 80)
    print(pkg)
    print("=" * 80)

    all_imports: list[str] = []
    for name in dir(pkg):
        submodule = getattr(pkg, name)
        if isinstance(submodule, ModuleType):
            # print(submodule)
            all = getattr(submodule, "__all__", None)
            all = cast(Union[tuple[str], None], all)
            if all is not None:
                if not submodule.__name__.startswith(pkg.__name__ + "."):
                    continue
                    # print(
                    #     f"{name}: {submodule.__name__} does not appear to be a submodule of {pkg.__name__}"
                    # )

                all_imports.extend(all)

                submodule_relname = submodule.__name__.removeprefix(pkg.__name__)
                out = f"from {submodule_relname} import " + ", ".join(all)
                print(out)

    # print(f"__all__ = {tuple(all_imports)}")


gen_init_py("shiny")
gen_init_py("shiny.reactive")
gen_init_py("shiny.render")
gen_init_py("shiny.session")
gen_init_py("shiny.ui")



import seaborn as sns

# Import data from shared.py
from shared import df

from shiny.express import input, render, ui

ui.page_opts(title="Shiny navigation components")

ui.nav_spacer()  # Push the navbar items to the right

footer = ui.input_select(
    "var", "Select variable", choices=["bill_length_mm", "body_mass_g"]
)

with ui.nav_panel("Page 1"):
    with ui.navset_card_underline(title="Penguins data", footer=footer):
        with ui.nav_panel("Plot"):

            @render.plot
            def hist():
                p = sns.histplot(
                    df, x=input.var(), facecolor="#007bc2", edgecolor="white"
                )
                return p.set(xlabel=None)

        with ui.nav_panel("Table"):

            @render.data_frame
            def data():
                return df[["species", "island", input.var()]]


with ui.nav_panel("Page 2"):
    "This is the second 'page'."



from pathlib import Path

from shiny.express import ui

js_file = Path(__file__).parent / "js" / "app.js"

"If you see this page before 'OK'-ing the alert box, something went wrong"

ui.include_js(js_file)



import pytest
from playwright.sync_api import Page
from utils.deploy_utils import (
    local_deploys_app_url_fixture,
    reruns,
    reruns_delay,
    skip_if_not_chrome,
)

from shiny.playwright import controller

app_url = local_deploys_app_url_fixture("express_page_fluid")


@skip_if_not_chrome
@pytest.mark.flaky(reruns=reruns, reruns_delay=reruns_delay)
def test_express_page_fluid(page: Page, app_url: str) -> None:
    page.goto(app_url)

    card = controller.Card(page, "card")
    output_txt = controller.OutputTextVerbatim(page, "txt")
    output_txt.expect_value("50")
    bounding_box = card.loc.bounding_box()
    assert bounding_box is not None
    assert bounding_box["height"] < 300



from shiny import App, render, ui

app_ui = ui.page_fluid(
    ui.navset_underline(
        ui.nav_panel("A", "Panel A content"),
        ui.nav_spacer(),
        ui.nav_panel("B", "Panel B content"),
        ui.nav_panel("C", "Panel C content"),
        id="selected_navset_underline",
    ),
    ui.h5("Selected:"),
    ui.output_code("selected"),
)


def server(input, output, session):
    @render.code
    def selected():
        return input.selected_navset_underline()


app = App(app_ui, server)



from __future__ import annotations

from typing import Optional

from htmltools import TagAttrs, TagAttrValue, TagChild, TagFunction, Tagifiable, tags

from ..._deprecated import warn_deprecated
from ...types import MISSING, MISSING_TYPE
from ...ui import CardItem
from ...ui import card as main_card
from ...ui._card import card_body as main_card_body  # FIXME after #1506
from ...ui.css import CssUnit


def card(
    *args: TagChild | TagAttrs | CardItem,
    full_screen: bool = False,
    height: Optional[CssUnit] = None,
    max_height: Optional[CssUnit] = None,
    min_height: Optional[CssUnit] = None,
    fill: bool = True,
    class_: Optional[str] = None,
    wrapper: MISSING_TYPE = MISSING,
    **kwargs: TagAttrValue,
):
    """Deprecated. Please use :func:`~shiny.ui.card()` instead."""
    warn_deprecated(
        "`shiny.experimental.ui.card()` was deprecated in shiny 1.0.0. "
        "Card components are now available in the main shiny namespace. "
        "Please use `shiny.ui.card()` instead."
    )
    return main_card(
        *args,
        full_screen=full_screen,
        height=height,
        max_height=max_height,
        min_height=min_height,
        fill=fill,
        class_=class_,
        id=None,
        **kwargs,
    )


def card_body(
    *args: TagChild | TagAttrs,
    fillable: bool = True,
    min_height: Optional[CssUnit] = None,
    max_height: Optional[CssUnit] = None,
    max_height_full_screen: Optional[CssUnit] | MISSING_TYPE = MISSING,
    height: Optional[CssUnit] = None,
    padding: Optional[CssUnit | list[CssUnit]] = None,
    gap: Optional[CssUnit] = None,
    fill: bool = True,
    class_: Optional[str] = None,
    **kwargs: TagAttrValue,
):
    """Deprecated. Please use :func:`~shiny.ui.card_body()` instead."""
    warn_deprecated(
        "`shiny.experimental.ui.card_body()` was deprecated in shiny 1.0.0. "
        "Card components are now available in the main shiny namespace. "
        "Please use `shiny.ui.card_body()` instead."
    )
    return main_card_body(
        *args,
        fillable=fillable,
        min_height=min_height,
        max_height=max_height,
        max_height_full_screen=max_height_full_screen,
        height=height,
        padding=padding,
        gap=gap,
        fill=fill,
        class_=class_,
        **kwargs,
    )


def card_title(
    *args: TagChild | TagAttrs,
    container: TagFunction = tags.h5,
    **kwargs: TagAttrValue,
) -> Tagifiable:
    """
    Deprecated. Please use :func:`~shiny.ui.tags.h5()` instead.

    :func:`~shiny.experimental.ui.card_title` creates a general container for the "title" of
    a :func:`~shiny.ui.card`. This component is designed
    to be provided as a direct child to :func:`~shiny.ui.card`.

    Parameters
    ----------
    *args
        Contents to appear in the card's title, or tag attributes to pass to the
        resolved :class:`~htmltools.Tag` object.
    container
        Method for the returned :class:`~htmltools.Tag` object. Defaults to
        :func:`~shiny.ui.tags.h5`.
    **kwargs
        Additional HTML attributes for the returned :class:`~htmltools.Tag` object.

    Returns
    -------
    :
        An :class:`~htmltools.Tag` object.

    See Also
    --------
    * :func:`~shiny.ui.card` for creating a card component.
    * :func:`~shiny.ui.card_header` for creating a header within a card.
    * :func:`~shiny.experimental.ui.card_body` for putting content inside a card.
    * :func:`~shiny.ui.card_footer` for creating a footer within a card.
    * :func:`~shiny.experimental.ui.card_image` for adding an image to a card.
    """
    warn_deprecated(
        "`shiny.experimental.ui.card_title()` was deprecated in shiny 1.0.0 "
        "and will be removed in a future version of shiny. "
        "Please use `shiny.ui.tags.h5()` instead."
    )
    return container(*args, **kwargs)


__all__ = (
    "card",
    "card_body",
)



from shiny import App, Inputs, Outputs, Session, module, render, ui


@module.ui
def ui_card():
    return ui.card(
        ui.card_header("Hello, card!"),
        "A regular full-screenable card.",
        id="thing",
        full_screen=True,
    )


@module.ui
def ui_value_box():
    return ui.value_box(
        "Hello, value box!",
        "$1,234,567",
        id="thing",
        full_screen=True,
        showcase=ui.span("$", class_="fs-1"),
    )


@module.server
def card_server(input: Inputs, output: Outputs, session: Session):
    return input.thing_full_screen


app_ui = ui.page_fluid(
    ui_card("card"),
    ui_value_box("value_box"),
    ui.output_code("out_card"),
    ui.output_code("out_value_box"),
)


def server(input: Inputs):
    my_card = card_server("card")
    my_value_box = card_server("value_box")

    @render.code()
    def out_card():
        return my_card()

    @render.code()
    def out_value_box():
        return my_value_box()


app = App(app_ui, server)



from playwright.sync_api import Page

from shiny.playwright import controller
from shiny.run import ShinyAppProc


def test_input_password_kitchen(page: Page, local_app: ShinyAppProc) -> None:
    page.goto(local_app.url)

    default = controller.InputPassword(page, "default")
    default.expect_label("Default password input")
    default.expect_value("")
    controller.OutputCode(page, "default_txt").expect_value("")

    placeholder = controller.InputPassword(page, "placeholder")
    placeholder.expect_placeholder("Enter password")
    placeholder.expect_value("")

    width = controller.InputPassword(page, "width")
    width.expect_width("200px")

    value = controller.InputPassword(page, "value")
    value.expect_value("secret123")
    controller.OutputCode(page, "value_txt").expect_value("secret123")
    value.set("secret456")
    value.expect_value("secret456")
    controller.OutputCode(page, "value_txt").expect_value("secret456")



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, Optional, TypeVar

from htmltools import Tag, TagAttrs, TagAttrValue, TagChild, css, tags

from .._docstring import add_example
from .._namespaces import resolve_id_or_none
from .._utils import drop_none, private_random_id
from ..session import require_active_session
from ..types import MISSING, MISSING_TYPE
from ._html_deps_shinyverse import components_dependencies
from ._tag import consolidate_attrs
from .css._css_unit import CssUnit, as_css_unit

if TYPE_CHECKING:
    from ..session import Session

__all__ = (
    "accordion",
    "accordion_panel",
    "insert_accordion_panel",
    "remove_accordion_panel",
    "update_accordion",
    "update_accordion_panel",
)


class AccordionPanel:
    """
    The internal class used to represent an accordion panel.

    This class is used to represent an accordion panel. It is not intended to be
    instantiated directly. Instead, use :func:`~shiny.ui.accordion_panel`.

    Parameters
    ----------
    *args
        Contents to appear in the accordion panel body, or tag attributes that are supplied to the
        returned :class:`~htmltools.Tag` object.
    data_value
        A character string that uniquely identifies this panel.
    icon
        A :class:`~htmltools.Tag` which is positioned just before the `title`.
    title
        A title to appear in the :func:`~shiny.ui.accordion_panel`'s header.
    id
        A unique id for this panel.
    **kwargs
        Tag attributes to the `accordion-body` div Tag.

    See Also
    --------
    * :func:`~shiny.ui.accordion`
    * :func:`~shiny.ui.update_accordion`
    * :func:`~shiny.ui.insert_accordion_panel`
    * :func:`~shiny.ui.remove_accordion_panel`
    * :func:`~shiny.ui.update_accordion_panel`
    """

    _args: tuple[TagChild | TagAttrs, ...]
    _kwargs: dict[str, TagAttrValue]

    _data_value: str  # Read within `accordion()`
    _icon: TagChild | None
    _title: TagChild | None
    _id: str | None

    _accordion_id: str | MISSING_TYPE  # Set within `accordion()`
    _is_open: bool  # Set within `accordion()`
    _is_multiple: bool  # Set within `accordion()`

    def __init__(
        self,
        *args: TagChild | TagAttrs,
        data_value: str,
        icon: TagChild | None,
        title: TagChild | None,
        id: str | None,
        **kwargs: TagAttrValue,
    ):
        self._args = args
        self._data_value = data_value
        self._icon = icon
        self._title = title
        self._id = id
        self._accordion_id = MISSING
        self._kwargs = kwargs
        self._is_multiple = False
        self._is_open = True

    def resolve(self) -> Tag:
        """
        Resolve the :class:`~shiny.ui.AccordionPanel` into a
        :class:`~htmltools.Tag`.

        Returns
        -------
        :
            A :class:`~htmltools.Tag` object representing the
            :class:`~shiny.ui.AccordionPanel`.
        """
        btn_attrs: TagAttrs = {}
        if self._is_open:
            btn_attrs["aria-expanded"] = "true"
        else:
            btn_attrs["class"] = "collapsed"
            btn_attrs["aria-expanded"] = "false"

        btn = tags.button(
            {
                "class": "accordion-button",
                "type": "button",
                "data-bs-toggle": "collapse",
                "data-bs-target": f"#{self._id}",
                "aria-controls": self._id,
            },
            btn_attrs,
            # Always include an .accordion-icon container to simplify update_accordion_panel() logic
            tags.div({"class": "accordion-icon"}, self._icon),
            tags.div({"class": "accordion-title"}, self._title),
        )

        attrs, children = consolidate_attrs(*self._args, **self._kwargs)

        if isinstance(self._accordion_id, MISSING_TYPE):
            raise RuntimeError(
                "Internal error: _accordion_id not set. Did you add your AccordionPanel an `accordion()`?"
            )

        return tags.div(
            {
                "class": "accordion-item",
                "data-value": self._data_value,
            },
            # Use a <span.h2> instead of <h2> so that it doesn't get included in rmd/pkgdown/qmd TOC
            # TODO-bslib: can we provide a way to put more stuff in the header? Like maybe some right-aligned controls?
            tags.span(
                {"class": "accordion-header h2"},
                btn,
            ),
            tags.div(
                {
                    "id": self._id,
                    "class": "accordion-collapse collapse",
                },
                {"class": "show"} if self._is_open else None,
                (
                    {"data-bs-parent": f"#{self._accordion_id}"}
                    if not self._is_multiple
                    else None
                ),
                tags.div(
                    {"class": "accordion-body"},
                    attrs,
                    children,
                ),
            ),
        )

    def tagify(self) -> Tag:
        """
        Resolve the :class:`~shiny.ui.AccordionPanel` into a
        :class:`~htmltools.Tag`.

        Returns
        -------
        :
            A tagified `resolve()`d value.
        """
        return self.resolve().tagify()


@add_example()
def accordion(
    *args: AccordionPanel | TagAttrs,
    id: Optional[str] = None,
    open: Optional[bool | str | list[str]] = None,
    multiple: bool = True,
    class_: Optional[str] = None,
    width: Optional[CssUnit] = None,
    height: Optional[CssUnit] = None,
    **kwargs: TagAttrValue,
) -> Tag:
    """
    Create a vertically collapsing accordion.

    Parameters
    ----------
    *args
        :class:`~shiny.ui.AccordionPanel` objects returned from
        :func:`~shiny.ui.accordion_panel`. Or tag attributes that are
        supplied to the returned :class:`~htmltools.Tag` object.
    id
        If provided, you can use `input.id()` in your server logic to determine which of
        the :func:`~shiny.ui.accordion_panel`s are currently active. The
        value will correspond to the :func:`~shiny.ui.accordion_panel`'s
        `value` argument.
    open
        A list of :func:`~shiny.ui.accordion_panel` values to open (i.e.,
        show) by default. The default value of `None` will open the first
        :func:`~shiny.ui.accordion_panel`. Use a value of `True` to open
        all (or `False` to open none) of the items. It's only possible to open more than
        one panel when `multiple=True`.
    multiple
        Whether multiple :func:`~shiny.ui.accordion_panel` can be open at
        once.
    class_
        Additional CSS classes to include on the accordion div.
    width
        Any valid CSS unit; for example, height="100%".
    height
        Any valid CSS unit; for example, height="100%".
    **kwargs
        Attributes to this tag.

    Returns
    -------
    :
        Accordion panel Tag object.


    References
    ----------
    [Bootstrap Accordion](https://getbootstrap.com/docs/5.3/components/accordion/)

    See Also
    --------
    * :func:`~shiny.ui.accordion_panel`
    * :func:`~shiny.ui.update_accordion`
    * :func:`~shiny.ui.insert_accordion_panel`
    * :func:`~shiny.ui.remove_accordion_panel`
    * :func:`~shiny.ui.update_accordion_panel`
    """

    # TODO-bookmarking: Restore input here
    # open = restore_input(id = id, default = open)

    attrs, panels = consolidate_attrs(*args, class_=class_, **kwargs)
    for panel in panels:
        if not isinstance(panel, AccordionPanel):
            raise TypeError(
                "All `accordion(*args)` must be of type `AccordionPanel` which can be created using `accordion_panel()`"
            )

    is_open: list[bool] = []
    if open is None:
        is_open = [False for _ in panels]
    elif isinstance(open, bool):
        is_open = [open for _ in panels]
    else:
        if not isinstance(open, list):
            open = [open]
        #
        is_open = [panel._data_value in open for panel in panels]

    # Open the first panel by default
    if open is not False and len(is_open) > 0 and not any(is_open):
        is_open[0] = True

    if (not multiple) and sum(is_open) > 1:
        raise ValueError("Can't select more than one panel when `multiple = False`")

    # Since multiple=False requires an id, we always include one,
    # but only create a binding when it is provided
    binding_class_value: TagAttrs | None = None
    if id is None:
        id = private_random_id("bslib_accordion")
        binding_class_value = None
    else:
        binding_class_value = {"class": "bslib-accordion-input"}

    accordion_id = resolve_id_or_none(id)
    for panel, open in zip(panels, is_open):
        panel._accordion_id = accordion_id
        panel._is_multiple = multiple
        panel._is_open = open

    panel_tags = [panel.resolve() for panel in panels]

    tag = tags.div(
        {
            "id": accordion_id,
            "class": "accordion",
            "style": css(width=as_css_unit(width), height=as_css_unit(height)),
        },
        # just for ease of identifying autoclosing client-side
        {"class": "autoclose"} if not multiple else None,
        binding_class_value,
        components_dependencies(),
        attrs,
        *panel_tags,
    )
    return tag


@add_example()
def accordion_panel(
    title: TagChild,
    *args: TagChild | TagAttrs,
    value: Optional[str] | MISSING_TYPE = MISSING,
    icon: Optional[TagChild] = None,
    **kwargs: TagAttrValue,
) -> AccordionPanel:
    """
    Single accordion panel.

    Parameters
    ----------
    title
        A title to appear in the :func:`~shiny.ui.accordion_panel`'s header.
    *args
        Contents to the accordion panel body. Or tag attributes that are supplied to the
        returned :class:`~htmltools.Tag` object.
    value
        A character string that uniquely identifies this panel. If `MISSING`, the
        `title` will be used.
    icon
        A :class:`~htmltools.Tag` which is positioned just before the `title`.
    **kwargs
        Tag attributes to the `accordion-body` div Tag.

    Returns
    -------
    :
        `AccordionPanel` object.


    References
    ----------
    [Bootstrap Accordion](https://getbootstrap.com/docs/5.3/components/accordion/)

    See Also
    --------
    * :func:`~shiny.ui.accordion`
    * :func:`~shiny.ui.update_accordion`
    * :func:`~shiny.ui.insert_accordion_panel`
    * :func:`~shiny.ui.remove_accordion_panel`
    * :func:`~shiny.ui.update_accordion_panel`
    """

    if value is MISSING:
        if isinstance(title, str):
            value = title
        else:
            raise ValueError("If `title` is not a string, `value` must be provided")
        value = title
    if not isinstance(value, str):
        raise TypeError("`value` must be a string")

    id = private_random_id("bslib_accordion_panel")

    return AccordionPanel(
        *args,
        data_value=value,
        icon=icon,
        title=title,
        id=id,
        **kwargs,
    )


# Send message before the next flush since things like remove/insert may
# remove/create input/output values. Also do this for set/open/close since,
# you might want to open a panel after inserting it.
def _send_panel_message(
    id: str,
    session: Session | None,
    **kwargs: object,
) -> None:
    message = drop_none(kwargs)
    session = require_active_session(session)
    session.on_flush(lambda: session.send_input_message(id, message), once=True)


def _accordion_panel_action(
    *,
    id: str,
    method: str,
    values: bool | str | list[str],
    session: Session | None,
) -> None:
    if not isinstance(values, bool):
        if not isinstance(values, list):
            values = [values]
        _assert_list_str(values)

    _send_panel_message(
        id,
        session,
        method=method,
        values=values,
    )


@add_example()
def update_accordion(
    id: str,
    *,
    show: bool | str | list[str],
    session: Optional[Session] = None,
) -> None:
    """
    Dynamically set accordions' states.

    Dynamically (i.e., programmatically) update/modify :func:`~shiny.ui.accordion`s in a Shiny app.
    These functions require an `id` to be provided to the :func:`~shiny.ui.accordion` and must also be
    called within an active Shiny session.

    Parameters
    ----------
    id
        A string that matches an existing :func:`~shiny.ui.accordion`'s `id`.
    show
        Either a string or list of strings (used to identify particular
        :func:`~shiny.ui.accordion_panel`(s) by their `value`) or a `bool` to set the state of all
        panels.
    session
        A Shiny session object (the default should almost always be used).

    References
    ----------
    [Bootstrap Accordion](https://getbootstrap.com/docs/5.3/components/accordion/)

    See Also
    --------
    * :func:`~shiny.ui.accordion`
    * :func:`~shiny.ui.accordion_panel`
    * :func:`~shiny.ui.insert_accordion_panel`
    * :func:`~shiny.ui.remove_accordion_panel`
    * :func:`~shiny.ui.update_accordion_panel`
    """
    if show is False:
        show_val = []
    else:
        show_val = show
    _accordion_panel_action(id=id, method="set", values=show_val, session=session)


@add_example()
def insert_accordion_panel(
    id: str,
    panel: AccordionPanel,
    target: Optional[str] = None,
    position: Literal["after", "before"] = "after",
    session: Optional[Session] = None,
) -> None:
    """
    Insert an :func:`~shiny.ui.accordion_panel`.

    Parameters
    ----------
    id
        A string that matches an existing :func:`~shiny.ui.accordion`'s `id`.
    panel
        An :func:`~shiny.ui.accordion_panel` object to insert.
    target
        The `value` of an existing panel to insert next to.
    position
        Should `panel` be added before or after the target? When `target=None`,
        `"after"` will append after the last panel and `"before"` will prepend before
        the first panel.
    session
        A Shiny session object (the default should almost always be used).

    References
    ----------
    [Bootstrap Accordion](https://getbootstrap.com/docs/5.3/components/accordion/)

    See Also
    --------
    * :func:`~shiny.ui.accordion`
    * :func:`~shiny.ui.accordion_panel`
    * :func:`~shiny.ui.update_accordion`
    * :func:`~shiny.ui.remove_accordion_panel`
    * :func:`~shiny.ui.update_accordion_panel`
    """

    if position not in ("after", "before"):
        raise ValueError("`position` must be either 'after' or 'before'")
    session = require_active_session(session)
    # Add accordion ID to panel; Used when `accordion(multiple=False)`
    panel._accordion_id = id

    _send_panel_message(
        id,
        session,
        method="insert",
        panel=session._process_ui(panel.resolve()),
        target=None if target is None else _assert_str(target),
        position=position,
    )


@add_example()
def remove_accordion_panel(
    id: str,
    target: str | list[str],
    session: Optional[Session] = None,
) -> None:
    """
    Remove an :func:`~shiny.ui.accordion_panel`.

    Parameters
    ----------
    id
        A string that matches an existing :func:`~shiny.ui.accordion`'s `id`.
    target
        The `value` of an existing panel to remove.
    session
        A Shiny session object (the default should almost always be used).

    References
    ----------
    [Bootstrap Accordion](https://getbootstrap.com/docs/5.3/components/accordion/)

    See Also
    --------
    * :func:`~shiny.ui.accordion`
    * :func:`~shiny.ui.accordion_panel`
    * :func:`~shiny.ui.update_accordion`
    * :func:`~shiny.ui.insert_accordion_panel`
    * :func:`~shiny.ui.update_accordion_panel`
    """
    if not isinstance(target, list):
        target = [target]

    _send_panel_message(
        id,
        session,
        method="remove",
        target=_assert_list_str(target),
    )


T = TypeVar("T")


def _missing_none_x(x: T | None | MISSING_TYPE) -> T | Literal[""] | None:
    if isinstance(x, MISSING_TYPE):
        return None
    if x is None:
        return ""
    return x


@add_example()
def update_accordion_panel(
    id: str,
    target: str,
    *body: TagChild,
    title: TagChild | None | MISSING_TYPE = MISSING,
    value: str | None | MISSING_TYPE = MISSING,
    icon: TagChild | None | MISSING_TYPE = MISSING,
    show: Optional[bool] = None,
    session: Optional[Session] = None,
) -> None:
    """
    Dynamically update accordion panel contents.

    Dynamically (i.e., programmatically) update/modify :func:`~shiny.ui.accordion` panels in a Shiny app.
    These functions require an `id` to be provided to the :func:`~shiny.ui.accordion` and must also be
    called within an active Shiny session.

    Parameters
    ----------
    id
        A string that matches an existing :func:`~shiny.ui.accordion`'s `id`.
    target
        The `value` of an existing panel to update.
    *body
        If provided, the new body contents of the panel.
    title
        If not missing, the new title of the panel.
    value
        If not missing, the new value of the panel.
    icon
        If not missing, the new icon of the panel.
    session
        A Shiny session object (the default should almost always be used).

    References
    ----------
    [Bootstrap Accordion](https://getbootstrap.com/docs/5.3/components/accordion/)

    See Also
    --------
    * :func:`~shiny.ui.accordion`
    * :func:`~shiny.ui.accordion_panel`
    * :func:`~shiny.ui.update_accordion`
    * :func:`~shiny.ui.insert_accordion_panel`
    * :func:`~shiny.ui.remove_accordion_panel`
    """

    session = require_active_session(session)

    # If `show` is given, then we need to open/close the targeted panel
    # Perform before changing `value` at the same time.
    if show is not None:
        _accordion_panel_action(
            id=id,
            method="open" if bool(show) else "close",
            values=[target],
            session=session,
        )

    title = _missing_none_x(title)
    value = _missing_none_x(value)
    icon = _missing_none_x(icon)
    _send_panel_message(
        id,
        session,
        method="update",
        target=_assert_str(target),
        value=None if value is None else _assert_str(value),
        body=None if len(body) == 0 else session._process_ui(body),
        title=None if title is None else session._process_ui(title),
        icon=None if icon is None else session._process_ui(icon),
    )


def _assert_str(x: str) -> str:
    if not isinstance(x, str):
        raise TypeError(f"Expected str, got {type(x)}")
    return x


def _assert_list_str(x: list[str]) -> list[str]:
    if not isinstance(x, list):
        raise TypeError(f"Expected list, got {type(x)}")
    for i, x_i in enumerate(x):
        if not isinstance(x_i, str):
            raise TypeError(f"Expected str in x[{i}], got {type(x_i)}")
    return x
